#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2021 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests the ppa command-line script."""

import os
import io
import sys
import types

import importlib.machinery
import argparse
from unittest.mock import patch

import pytest

SCRIPT_NAME = "ppa"
BASE_PATH = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), ".."))
sys.path.insert(0, BASE_PATH)

from ppa.ppa import Ppa
from ppa.constants import (
    ARCHES_PPA_ALL,
    ARCHES_PPA_DEFAULT
)
from tests.helpers import (
    LpServiceMock,
    PublicationMock
)

if '.pybuild' in BASE_PATH:
    # pylint: disable-next=invalid-name
    python_version = '.'.join([str(v) for v in sys.version_info[0:2]])
    scripts_path = os.path.join(
        BASE_PATH.replace(f'/.pybuild/cpython3_{python_version}/', '/'),
        f'scripts-{python_version}'
    )
else:
    scripts_path = os.path.join(BASE_PATH, 'scripts')

script_path = os.path.join(scripts_path, SCRIPT_NAME)
loader = importlib.machinery.SourceFileLoader(SCRIPT_NAME, script_path)
script = types.ModuleType(loader.name)
loader.exec_module(script)


@pytest.fixture
def fake_config():
    return {
        'ppa_name': 'testing',
        'owner_name': 'me',
        'wait_seconds': 0.1,
        'quiet': True
    }


@pytest.fixture
def fake_source_package():
    # TODO: Implement
    return {
        'name': 'test-source',
        'version': '1.0-2ubuntu3',
        'binaries': [
            {'name': 'foo'},
            {'name': 'bar'}
        ]
    }


@pytest.fixture
def fake_binary_package():
    # TODO: Implement
    return {
        'name': 'test-binary',
        'source': 'test-source',
    }


@pytest.fixture
def fake_build():
    # TODO: Implement, including build errors
    return {
        'binary': 'foo',
        'status': 'Failed',
    }


@pytest.fixture
def fake_bug_report():
    # TODO: Implement
    return {
        'id': 1234,
        'title': 'foobar',
        'desc': 'baz',
    }


def test_create_arg_parser():
    """
    Checks that the main argument processor is created properly.
    It must support the top level options as well as the expected
    set of subparsers.

    Note we don't test --help or --version since these are built-ins
    from argparse so we don't control their behavior.  They also both
    exit when done which is not conducive to testing.
    """
    parser = script.create_arg_parser()

    # Check command is recognized
    args = parser.parse_args(['show', 'test-ppa'])
    assert args.command == 'show'

    # Check -A, --credentials
    args = parser.parse_args(['-A', 'my-creds', 'show', 'test-ppa'])
    assert args.credentials_filename == "my-creds"
    args.credentials_filename = None
    args = parser.parse_args(['--credentials', 'my-creds', 'show', 'test-ppa'])
    assert args.credentials_filename == "my-creds"
    args.credentials_filename = None

    # Check -D, --debug
    args = parser.parse_args(['-D', 'show', 'test-ppa'])
    assert args.debug is True
    args.debug = None
    args = parser.parse_args(['--debug', 'show', 'test-ppa'])
    assert args.debug is True
    args.debug = None

    # Check -q, --dry-run
    args = parser.parse_args(['--dry-run', 'show', 'test-ppa'])
    assert args.dry_run is True
    args.dry_run = None

    # Check -v, --verbose
    args = parser.parse_args(['-v', 'show', 'test-ppa'])
    assert args.verbose is True
    args.verbose = None
    args = parser.parse_args(['--verbose', 'show', 'test-ppa'])
    assert args.verbose is True
    args.verbose = None

    # Check -q, --quiet
    args = parser.parse_args(['-q', 'show', 'test-ppa'])
    assert args.quiet is True
    args.quiet = None
    args = parser.parse_args(['--quiet', 'show', 'test-ppa'])
    assert args.quiet is True
    args.quiet = None

    # Verify all expected subparsers are present
    subparsers_actions = [
        action for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)]
    subparsers = []
    for subparsers_action in subparsers_actions:
        for choice, _ in subparsers_action.choices.items():
            subparsers.append(choice)
    assert subparsers == [
        'create',
        'credentials',
        'desc',
        'destroy',
        'list',
        'set',
        'show',
        'tests',
        'wait'
    ]


@pytest.mark.parametrize('command', ['create', 'set'])
def test_create_arg_parser_basic_config(command):
    """Checks argument parsing for the basic PPA config options.

    This test covers the set of options used by 'create' and 'set' for
    configuring various properties of the PPA's behaviors, such as
    dependencies, publication policy, access control, etc.  It does not
    cover settings that require Launchpad administrator involvement.

    The testing checks only that the options are being received and
    registered as expected, and does not cover the processing of the
    inputs nor the actual underlying functionality.
    """
    parser = script.create_arg_parser()

    # Check command and ppa_name
    args = parser.parse_args([command, 'test-ppa'])
    assert args.command == command
    assert args.ppa_name == 'test-ppa'

    # Check that command args can come before or after the ppa name
    args = parser.parse_args([command, 'test-ppa', '-a', 'x'])
    assert args.architectures == 'x'
    args.architectures = None
    args = parser.parse_args([command, '-a', 'x', 'test-ppa'])
    assert args.architectures == 'x'
    args.architectures = None

    # Check --all-arches and --default-arches
    args = parser.parse_args([command, 'test-ppa', '--all-arches'])
    assert args.architectures == ','.join(ARCHES_PPA_ALL)
    args.architectures = None
    args = parser.parse_args([command, 'test-ppa', '--all-architectures'])
    assert args.architectures == ','.join(ARCHES_PPA_ALL)
    args.architectures = None
    args = parser.parse_args([command, 'test-ppa', '--default-arches'])
    assert args.architectures == ','.join(ARCHES_PPA_DEFAULT)
    args.architectures = None
    args = parser.parse_args([command, 'test-ppa', '--default-architectures'])
    assert args.architectures == ','.join(ARCHES_PPA_DEFAULT)
    args.architectures = None

    # Check -a, --arch, --arches, --architectures
    args = parser.parse_args([command, 'test-ppa', '-a', 'x'])
    assert args.architectures == 'x'
    args.architectures = None
    args = parser.parse_args([command, 'test-ppa', '--arch', 'x'])
    assert args.architectures == 'x'
    args.architectures = None
    args = parser.parse_args([command, 'test-ppa', '--arches', 'x'])
    assert args.architectures == 'x'
    args.architectures = None
    args = parser.parse_args([command, 'test-ppa', '--architectures', 'a,b,c'])
    assert args.architectures == 'a,b,c'
    args.architectures = None

    # Check --displayname
    args = parser.parse_args([command, 'test-ppa', '--displayname', 'x'])
    assert args.displayname == 'x'
    args.displayname = None

    # Check --description
    args = parser.parse_args([command, 'test-ppa', '--description', 'x'])
    assert args.description == 'x'
    args.description = None

    # Check --public / -P|--private
    args = parser.parse_args([command, 'test-ppa', '--public'])
    assert args.private is False
    args.public = None
    args = parser.parse_args([command, 'test-ppa', '--private'])
    assert args.private is True
    args.private = None
    args = parser.parse_args([command, 'test-ppa', '-P'])
    assert args.private is True
    args.private = None

    # Check --ppa-dependencies <PPA[,...]>
    args = parser.parse_args([command, 'test-ppa', '--ppa-dependencies', 'a,b,c'])
    assert args.ppa_dependencies == "a,b,c"
    args.ppa_dependencies = None
    args = parser.parse_args([command, 'test-ppa', '--ppa-depends', 'a,b,c'])
    assert args.ppa_dependencies == "a,b,c"
    args.ppa_dependencies = None

    # Check --publish
    args = parser.parse_args([command, 'test-ppa', '--publish'])
    assert args.publish is True
    args.publish = None
    args = parser.parse_args([command, 'test-ppa', '--no-publish'])
    assert args.publish is False
    args.publish = None


def test_create_arg_parser_create():
    """Checks argument parsing for the 'create' command.

    Most of the create command's args are covered by
    test_create_arg_parser_basic_config(), this just verifies
    the few that aren't.
    """
    parser = script.create_arg_parser()
    command = 'create'

    # Check ppa_name
    args = parser.parse_args([command, 'test-ppa'])
    assert args.ppa_name == 'test-ppa'
    args = parser.parse_args([command, 'my-team/test-ppa'])
    assert args.ppa_name == 'my-team/test-ppa'
    args = parser.parse_args([command, 'ppa:my-team/test-ppa'])
    assert args.ppa_name == 'ppa:my-team/test-ppa'

    # Check --owner, --owner-name, --team, --team-name
    args = parser.parse_args([command, 'test-ppa', '--owner', 'x'])
    assert args.owner_name == 'x'
    args.owner_name = None
    args = parser.parse_args([command, 'test-ppa', '--owner-name', 'x'])
    assert args.owner_name == 'x'
    args.owner_name = None
    args = parser.parse_args([command, 'test-ppa', '--team', 'x'])
    assert args.owner_name == 'x'
    args.owner_name = None
    args = parser.parse_args([command, 'test-ppa', '--team-name', 'x'])
    assert args.owner_name == 'x'
    args.owner_name = None

    # Check --pocket
    args = parser.parse_args([command, 'test-ppa', '--pocket', 'proposed'])
    assert args.pocket == "proposed"
    args.pocket = None


def test_create_arg_parser_show():
    """Checks argument parsing for the 'show' command."""
    parser = script.create_arg_parser()
    command = 'show'

    # Check ppa_name
    args = parser.parse_args([command, 'test-ppa'])
    assert args.ppa_name == 'test-ppa'

    # Check -a, --arch, --arches, --architectures
    args = parser.parse_args([command, 'test-ppa', '-a', 'x'])
    assert args.architectures == 'x'
    args.architectures = None

    args = parser.parse_args([command, 'test-ppa', '--arch', 'x'])
    assert args.architectures == 'x'
    args.architectures = None
    args = parser.parse_args([command, 'test-ppa', '--arches', 'x'])
    assert args.architectures == 'x'
    args.architectures = None
    args = parser.parse_args([command, 'test-ppa', '--architectures', 'a,b,c'])
    assert args.architectures == 'a,b,c'
    args.architectures = None

    # Check -r, --release, --releases
    args = parser.parse_args([command, 'test-ppa', '-r', 'x'])
    assert args.releases == 'x'
    args.releases = None
    args = parser.parse_args([command, 'test-ppa', '--release', 'x'])
    assert args.releases == 'x'
    args.releases = None
    args = parser.parse_args([command, 'test-ppa', '--releases', 'x'])
    assert args.releases == 'x'
    args.releases = None
    args = parser.parse_args([command, 'test-ppa', '--releases', 'x,y,z'])
    assert args.releases == 'x,y,z'
    args.releases = None

    # Check -p, --package, --packages
    args = parser.parse_args([command, 'tests-ppa', '-p', 'x'])
    assert args.packages == 'x'
    args.packages = None
    args = parser.parse_args([command, 'tests-ppa', '--package', 'x'])
    assert args.packages == 'x'
    args.packages = None
    args = parser.parse_args([command, 'tests-ppa', '--packages', 'x'])
    assert args.packages == 'x'
    args.packages = None
    args = parser.parse_args([command, 'tests-ppa', '--packages', 'x,y,z'])
    assert args.packages == 'x,y,z'
    args.packages = None


def test_create_arg_parser_tests():
    """Checks argument parsing for the 'tests' command."""
    parser = script.create_arg_parser()
    command = 'tests'

    # Check ppa_name
    args = parser.parse_args([command, 'test-ppa'])
    assert args.ppa_name == 'test-ppa'

    # Check -a, --arch, --arches, --architectures
    args = parser.parse_args([command, 'test-ppa', '-a', 'x'])
    assert args.architectures == 'x'
    args.architectures = None

    args = parser.parse_args([command, 'test-ppa', '--arch', 'x'])
    assert args.architectures == 'x'
    args.architectures = None
    args = parser.parse_args([command, 'test-ppa', '--arches', 'x'])
    assert args.architectures == 'x'
    args.architectures = None
    args = parser.parse_args([command, 'test-ppa', '--architectures', 'a,b,c'])
    assert args.architectures == 'a,b,c'
    args.architectures = None

    # Check -r, --release, --releases
    args = parser.parse_args([command, 'test-ppa', '-r', 'x'])
    assert args.releases == 'x'
    args.releases = None
    args = parser.parse_args([command, 'test-ppa', '--release', 'x'])
    assert args.releases == 'x'
    args.releases = None
    args = parser.parse_args([command, 'test-ppa', '--releases', 'x'])
    assert args.releases == 'x'
    args.releases = None
    args = parser.parse_args([command, 'test-ppa', '--releases', 'x,y,z'])
    assert args.releases == 'x,y,z'
    args.releases = None

    # Check -p, --package, --packages
    args = parser.parse_args([command, 'tests-ppa', '-p', 'x'])
    assert args.packages == 'x'
    args.packages = None
    args = parser.parse_args([command, 'tests-ppa', '--package', 'x'])
    assert args.packages == 'x'
    args.packages = None
    args = parser.parse_args([command, 'tests-ppa', '--packages', 'x'])
    assert args.packages == 'x'
    args.packages = None
    args = parser.parse_args([command, 'tests-ppa', '--packages', 'x,y,z'])
    assert args.packages == 'x,y,z'
    args.packages = None

    # Check --show-urls, --show-url, -L
    args = parser.parse_args([command, 'tests-ppa'])
    assert args.show_urls is None
    args.show_urls = None
    args = parser.parse_args([command, 'tests-ppa', '--show-urls'])
    assert args.show_urls is True
    args.show_urls = None
    args = parser.parse_args([command, 'tests-ppa', '--show-url'])
    assert args.show_urls is True
    args.show_urls = None
    args = parser.parse_args([command, 'tests-ppa', '-L'])
    assert args.show_urls is True
    args.show_urls = None

    # Check --show-rdepends
    args = parser.parse_args([command, 'tests-ppa'])
    assert args.show_rdepends is None
    args.show_rdepends = None
    args = parser.parse_args([command, 'tests-ppa', '--show-rdepends'])
    assert args.show_rdepends is True
    args.show_rdepends = None


def test_create_arg_parser_wait():
    """Checks argument parsing for the 'wait' command."""
    parser = script.create_arg_parser()
    command = "wait"

    # Check ppa_name
    args = parser.parse_args([command, 'test-ppa'])
    assert args.ppa_name == 'test-ppa'

    # Check -l, --log
    args = parser.parse_args([command, 'test-ppa'])
    assert args.wait_logging is None
    args = parser.parse_args([command, 'test-ppa', '-l'])
    assert args.wait_logging is True
    args = parser.parse_args([command, 'test-ppa', '--log'])
    assert args.wait_logging is True


@pytest.mark.parametrize('command_line_options, expected_config', [
    pytest.param([], {}),
    (['show', 'ppa:aa/bb'], {'command': 'show', 'owner_name': 'aa', 'ppa_name': 'bb'}),
    (['show', 'aa/bb'], {'owner_name': 'aa', 'ppa_name': 'bb'}),
    (['show', 'bb'], {'owner_name': 'me', 'ppa_name': 'bb'}),
    (['--debug', 'show', 'ppa:aa/bb'], {'debug': True}),
    (['--dry-run', 'show', 'ppa:aa/bb'], {'dry_run': True}),
    (['--verbose', 'show', 'ppa:aa/bb'], {'verbose': True}),
    (['--quiet', 'show', 'ppa:aa/bb'], {'quiet': True}),
    (['create', 'ppa:aa/bb'], {'command': 'create', 'owner_name': 'aa', 'ppa_name': 'bb'}),
    (['create', 'aa/bb'], {'command': 'create', 'owner_name': 'aa', 'ppa_name': 'bb'}),
    (['create', 'bb', '--owner', 'aa'], {'command': 'create', 'owner_name': 'aa', 'ppa_name': 'bb'}),
    (['create', 'bb', '--owner-name', 'aa'], {'command': 'create', 'owner_name': 'aa', 'ppa_name': 'bb'}),
    (['create', 'bb', '--private'], {'command': 'create', 'private': True}),
    (['create', 'bb', '--public'], {'command': 'create', 'private': False}),
    (['create', 'bb', '--team', 'aa'], {'command': 'create', 'owner_name': 'aa', 'ppa_name': 'bb'}),
    (['create', 'bb', '--team-name', 'aa'], {'command': 'create', 'owner_name': 'aa', 'ppa_name': 'bb'}),
    ])
def test_create_config_from_args(command_line_options, expected_config):
    '''Checks creation of a config object from an argparser object.

    Prior tests cover the creation of proper args from the command line;
    this test relies on the already-tested argparse machinery to create
    various args to pass to create_config() in order to assure that the
    right config dict is generated in response.
    '''
    lp = LpServiceMock()
    parser = script.create_arg_parser()
    args = parser.parse_args(command_line_options)
    config = script.create_config(lp, args)

    for key, value in expected_config.items():
        assert key in config.keys()
        assert config[key] == value


@pytest.mark.parametrize('args, expected_exception', [
    # Bad command
    ([None], SystemExit),
    ([''], SystemExit),
    (['INVALID'], SystemExit),
    ([1], TypeError),

    # Bad ppa name
    (['show'], SystemExit),
    (['show', None], ValueError),
    (['show', ''], ValueError),
    (['show', 'INVALID'], ValueError),
    (['show', 1], TypeError),

    # Bad argument name
    (['--invalid', 'show', 'ppa:aa/bb'], SystemExit),
    ])
def test_create_config_from_args_error(args, expected_exception):
    '''Checks creation of a config object from an argparser object.'''
    lp = LpServiceMock()
    parser = script.create_arg_parser()

    with pytest.raises(expected_exception):
        args = parser.parse_args(args)
        script.create_config(lp, args)


@pytest.mark.parametrize('config, expected_config', [
    ({}, {}),
    ({'x': True}, {'x': True}),
    ({'ppa_dependencies': 'x,y,z'}, {'ppa_dependencies': ['x', 'y', 'z']}),
    ({'architectures': 'x,y,z'}, {'architectures': ['x', 'y', 'z']}),
    ({'releases': 'x,y,z'}, {'releases': ['x', 'y', 'z']}),
    ({'packages': 'x,y,z'}, {'packages': ['x', 'y', 'z']}),
    ])
def test_unpack_config(config, expected_config):
    '''Checks that processed config matches expectations.'''
    assert script.unpack_config(config) == expected_config


def test_create_lp_via_file(tmp_path):
    '''Checks loading of credentials provided by --credentials arg.'''
    expected_credentials = "[1]\na: 1\nb: 2\n"
    credentials_file = tmp_path / 'x'
    credentials_file.write_text(expected_credentials)

    lp = LpServiceMock()
    parser = script.create_arg_parser()
    args = parser.parse_args([
        '--credentials', str(credentials_file),
        'show', 'test-ppa',
    ])
    lp = script.create_lp('x', args)

    assert lp._credentials == expected_credentials


@patch.dict(os.environ, {"LP_CREDENTIALS": "[1]\na: 1\nb: 2\n"})
def test_create_lp_via_envvar(tmp_path):
    '''Checks loading of credentials provided by $LP_CREDENTIALS.'''
    lp = LpServiceMock()
    parser = script.create_arg_parser()
    args = parser.parse_args([
        'show', 'test-ppa',
    ])
    lp = script.create_lp('x', args)

    assert lp._credentials == os.environ['LP_CREDENTIALS']


@pytest.mark.parametrize('stdin, params, expected_ppa_config', [
    # Defaults
    (None, {}, {'description': ''}),

    # Overrides
    ('x', {}, {'description': 'x'}),
    (None, {'description': 'a'}, {'description': 'a'}),
    (None, {'ppa_name': 'a'}, {'displayname': 'a'}),
    (None, {'publish': True}, {'publish': True}),
    (None, {'publish': False}, {'publish': False}),
    (None, {'private': True}, {'private': True}),
    (None, {'private': False}, {'private': False}),
])
def test_command_create(fake_config, monkeypatch, stdin, params, expected_ppa_config):
    '''Checks create command produces a PPA with expected configuration.'''
    lp = LpServiceMock()
    monkeypatch.setattr("sys.stdin", io.StringIO(stdin))

    # Check success of the create command
    config = script.unpack_config({**fake_config, **params})
    print(config)
    assert script.command_create(lp, config) == 0

    # Retrieve the newly created PPA
    owner = lp.people[config['owner_name']]
    lp_ppa = owner.getPPAByName(config['ppa_name'])
    assert lp_ppa

    # Verify the expected items are present in the new PPA
    for key, value in expected_ppa_config.items():
        assert getattr(lp_ppa, key) == value


@pytest.mark.parametrize('params, owner_name', [
    # Defaults
    ({'owner_name': 'a', 'ppa_name': 'x'}, 'a'),
])
def test_command_create_with_owner(fake_config, monkeypatch, params, owner_name):
    '''Checks create command produces a PPA for a specified owner.'''
    lp = LpServiceMock()
    lp.launchpad.add_person(owner_name)
    monkeypatch.setattr("sys.stdin", io.StringIO('x'))

    # Check success of the create command
    config = script.unpack_config({**fake_config, **params})
    print(config)
    assert script.command_create(lp, config) == 0

    # Retrieve the newly created PPA
    owner = lp.people[owner_name]
    lp_ppa = owner.getPPAByName(config['ppa_name'])
    assert lp_ppa


@pytest.mark.parametrize('architectures, expected_processors', [
    (None, ARCHES_PPA_DEFAULT),
    ('a', ['a']),
    ('a,b,c', ['a', 'b', 'c']),
    ('a, b, c', ['a', 'b', 'c']),
    ('amd64,arm64,armhf,i386,powerpc,ppc64el,s390x',
     ["amd64", "arm64", "armhf", "i386", "powerpc", "ppc64el", "s390x"]),
])
def test_command_create_with_architectures(monkeypatch, fake_config, architectures, expected_processors):
    '''Checks that PPAs can be created with non-default architecture support.'''
    lp = LpServiceMock()
    config = script.unpack_config({**fake_config, **{'architectures': architectures}})
    monkeypatch.setattr("sys.stdin", io.StringIO('x'))
    assert script.command_create(lp, config) == 0

    # Retrieve the newly created PPA
    owner = lp.people[config['owner_name']]
    lp_ppa = owner.getPPAByName(config['ppa_name'])

    # Check processor architectures
    assert lp_ppa.processors
    assert type(lp_ppa.processors) is list
    assert [proc.name for proc in lp_ppa.processors] == expected_processors


@pytest.mark.parametrize('params, expected_filename', [
    ({}, 'credentials.oauth'),
    ({'credentials_filename': 'a'}, 'a'),
    ({'credentials_filename': '~/.x/creds.oauth'}, '~/.x/creds.oauth'),
])
def test_command_credentials(fake_config, params, expected_filename):
    """Check that credentials command saves LP creds to file."""
    lp = LpServiceMock()
    config = script.unpack_config({**fake_config, **params})

    assert script.command_credentials(lp, config) == 0
    lp.credentials.save_to_path.assert_called_once()


@pytest.mark.xfail(reason="Unimplemented")
def test_command_desc(fake_config):
    lp = LpServiceMock()
    assert script.command_desc(lp, fake_config) == 0
    # TODO: Assert that if --dry-run specified, there are no actual
    #   changes requested of launchpad
    # TODO: Verify the description gets set as expected


@pytest.mark.xfail(reason="Unimplemented")
def test_command_destroy(fake_config):
    lp = LpServiceMock()
    # TODO: Create a fake ppa to be destroyed
    assert script.command_destroy(lp, fake_config) == 0
    # TODO: Verify the ppa is requested to be deleted


@pytest.mark.xfail(reason="Unimplemented")
def test_command_list(fake_config):
    lp = LpServiceMock()
    # TODO: Create a fake ppa with contents to be listed
    assert script.command_list(lp, fake_config) == 0
    # TODO: Verify the ppa addresses get listed


@pytest.mark.xfail(reason="Unimplemented")
def test_command_exists(fake_config):
    lp = LpServiceMock()
    # TODO: Create fake ppa that exists
    assert script.command_exists(lp, fake_config) == 0
    # TODO: Verify this returns true when the ppa does exist


@pytest.mark.xfail(reason="Unimplemented")
def test_command_not_exists(fake_config):
    lp = LpServiceMock()
    # TODO: Verify this returns true when the ppa does not exist
    assert script.command_exists(lp, fake_config) == 1


@pytest.mark.parametrize('params, expected_ppa_config', [
    ({'displayname': 'a'}, {'displayname': 'a'}),
    ({'description': 'a'}, {'description': 'a'}),

    ({}, {'publish': True}),
    ({'publish': False}, {'publish': False}),
    ({'publish': True}, {'publish': True}),

    ({}, {'private': False}),
    ({'private': False}, {'private': False}),
    ({'private': True}, {'private': True}),
])
def test_command_set(fake_config, params, expected_ppa_config):
    '''Checks that the set command properly requests PPA configuration changes.'''
    lp = LpServiceMock()

    # Create a default PPA, for modification later
    owner = lp.people[fake_config['owner_name']]
    owner.createPPA(fake_config['ppa_name'], 'x', 'y')

    # Check success of the set command
    config = script.unpack_config({**fake_config, **params})
    assert script.command_set(lp, config)

    # Retrieve the PPA we created earlier
    lp_ppa = owner.getPPAByName(fake_config['ppa_name'])

    # Verify the expected items are present in the updated PPA
    for key, value in expected_ppa_config.items():
        assert getattr(lp_ppa, key) == value


@pytest.mark.parametrize('architectures, expected_processors', [
    (None, ARCHES_PPA_DEFAULT),
    ('a', ['a']),
    ('a,b,c', ['a', 'b', 'c']),
    ('a, b, c', ['a', 'b', 'c']),
    ('amd64,arm64,armhf,i386,powerpc,ppc64el,s390x',
     ["amd64", "arm64", "armhf", "i386", "powerpc", "ppc64el", "s390x"]),
])
def test_command_set_architectures(fake_config, architectures, expected_processors):
    '''Checks that existing PPAs can have their architectures changed.'''
    lp = LpServiceMock()

    # Create a default PPA, for modification later
    owner = lp.people[fake_config['owner_name']]
    owner.createPPA(fake_config['ppa_name'], 'x', 'y')

    # Check success of the set command
    config = script.unpack_config({**fake_config, **{'architectures': architectures}})
    assert script.command_set(lp, config)

    # Retrieve the PPA we created earlier
    lp_ppa = owner.getPPAByName(fake_config['ppa_name'])

    # Check processor architectures
    assert lp_ppa.processors
    assert type(lp_ppa.processors) is list
    assert [proc.name for proc in lp_ppa.processors] == expected_processors


@pytest.mark.parametrize('params, pubs, format, expected_in_stdout', [
    (
        # General behavior
        {
            'releases': None,
            'architectures': 'amd64',
            'show_urls': True
        },
        [('x', '1', 'Published', 'jammy')],
        'plain',
        'arch=amd64&trigger=x%2F1&ppa=me%2Ftesting'
    ),
    (
        # Specified release
        {
            'releases': 'focal',
            'architectures': 'amd64',
            'show_urls': True
        },
        [
            ('x', '3.2.1', 'Published', 'jammy'),
            ('x', '1.2.3', 'Published', 'focal'),
            ('x', '1.0.0', 'Published', 'bionic'),
        ],
        'plain',
        'arch=amd64&trigger=x%2F1.2.3&ppa=me%2Ftesting'
    ),
    (
        # Specified architectures can be comma-delimited strings
        {
            'releases': 'focal',
            'architectures': 'amd64,i386',
            'show_urls': True
        },
        [('x', '1.2.3', 'Published', 'focal')],
        'plain',
        'arch=i386&trigger=x%2F1.2.3&ppa=me%2Ftesting'
    ),
    (
        # Specified architectures can be provided as a list, too
        {
            'releases': 'focal',
            'architectures': ['amd64', 'i386', 's390x'],
            'show_urls': True
        },
        [('x', '1.2.3', 'Published', 'focal')],
        'plain',
        'arch=s390x&trigger=x%2F1.2.3&ppa=me%2Ftesting'
    ),
    (
        # Default behavior should include ESM releases (LP: #2038651)
        {
            'releases': None,
            'architectures': 'amd64',
            'show_urls': True
        },
        [
            ('x', '3.2.1', 'Published', 'jammy'),
            ('x', '1.2.3', 'Published', 'focal'),
            ('x', '1.0.0', 'Published', 'bionic'),
        ],
        'plain',
        'arch=amd64&trigger=x%2F1.0.0&ppa=me%2Ftesting'
    ),
])
@patch('urllib.request.urlopen')
@patch('ppa.io.open_url')
def test_command_show(urlopen_mock,
                      open_url_mock,
                      fake_config, capfd,
                      params, pubs, format, expected_in_stdout):
    '''Checks that the show command retrieves and displays correct results.'''
    lp = LpServiceMock()
    urlopen_mock.return_value = "{}"
    Ppa.get_autopkgtest_running = lambda x, y, z: []
    Ppa.get_autopkgtest_waiting = lambda x, y, z: []

    # Create a default PPA, for modification later
    owner = lp.people[fake_config['owner_name']]
    owner.createPPA(fake_config['ppa_name'], 'x', 'y')
    the_ppa = lp.me.getPPAByName(fake_config['ppa_name'])

    # Add some fake publications
    for pub in pubs:
        the_ppa.published_sources.append(PublicationMock(*pub))

    config = script.unpack_config({**fake_config, **params})
    assert script.command_show(lp, config) == 0
#    out, err = capfd.readouterr()
#    assert expected_in_stdout in out


@pytest.mark.parametrize('params, pubs, format, expected_in_stdout', [
    (
        # General behavior
        {
            'releases': None,
            'architectures': 'amd64',
            'show_urls': True
        },
        [('x', '1', 'Published', 'jammy')],
        'plain',
        'arch=amd64&trigger=x%2F1&ppa=me%2Ftesting'
    ),
    (
        # Specified release
        {
            'releases': 'focal',
            'architectures': 'amd64',
            'show_urls': True
        },
        [
            ('x', '3.2.1', 'Published', 'jammy'),
            ('x', '1.2.3', 'Published', 'focal'),
            ('x', '1.0.0', 'Published', 'bionic'),
        ],
        'plain',
        'arch=amd64&trigger=x%2F1.2.3&ppa=me%2Ftesting'
    ),
    (
        # Specified architectures can be comma-delimited strings
        {
            'releases': 'focal',
            'architectures': 'amd64,i386',
            'show_urls': True
        },
        [('x', '1.2.3', 'Published', 'focal')],
        'plain',
        'arch=i386&trigger=x%2F1.2.3&ppa=me%2Ftesting'
    ),
    (
        # Specified architectures can be provided as a list, too
        {
            'releases': 'focal',
            'architectures': ['amd64', 'i386', 's390x'],
            'show_urls': True
        },
        [('x', '1.2.3', 'Published', 'focal')],
        'plain',
        'arch=s390x&trigger=x%2F1.2.3&ppa=me%2Ftesting'
    ),
    (
        # Default behavior should include ESM releases (LP: #2038651)
        {
            'releases': None,
            'architectures': 'amd64',
            'show_urls': True
        },
        [
            ('x', '3.2.1', 'Published', 'jammy'),
            ('x', '1.2.3', 'Published', 'focal'),
            ('x', '1.0.0', 'Published', 'bionic'),
        ],
        'plain',
        'arch=amd64&trigger=x%2F1.0.0&ppa=me%2Ftesting'
    ),
])
@patch('urllib.request.urlopen')
@patch('ppa.io.open_url')
def test_command_tests(urlopen_mock,
                       open_url_mock,
                       fake_config, capfd,
                       params, pubs, format, expected_in_stdout):
    '''Checks that the tests command retrieves and displays correct results.'''
    lp = LpServiceMock()
    urlopen_mock.return_value = "{}"
    Ppa.get_autopkgtest_running = lambda x, y, z: []
    Ppa.get_autopkgtest_waiting = lambda x, y, z: []

    # Create a default PPA, for modification later
    owner = lp.people[fake_config['owner_name']]
    owner.createPPA(fake_config['ppa_name'], 'x', 'y')
    the_ppa = lp.me.getPPAByName(fake_config['ppa_name'])

    # Add some fake publications
    for pub in pubs:
        the_ppa.published_sources.append(PublicationMock(*pub))

    config = script.unpack_config({**fake_config, **params})
    assert script.command_tests(lp, config) == 0
    out, err = capfd.readouterr()
    assert expected_in_stdout in out


@pytest.mark.xfail(reason="Unimplemented")
def test_command_wait(fake_config):
    lp = LpServiceMock()
    # TODO: Set wait period to 1 sec
    assert script.command_wait(lp, fake_config) == 0



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2012 Bryce Harrington <bryce@bryceharrington.org>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from time import sleep
from subprocess import (Popen, PIPE)

from .debug import dbg
from .text import o2str


class ReturnCode(BaseException):
    def __init__(self, code, errors=None, output=None):
        self.code = code
        self.output = output
        if type(errors) in (list, tuple):
            self.errors = errors
        else:
            self.errors = [errors]

    def __str__(self):
        text = '\n'.join(self.errors)
        return f"{text}Returned error code {self.code}"


def shell(command, in_text=None):
    """Execute command in a shell, returns stdout, raises exception on error."""
    dbg(f"shell: {command}")
    if in_text:
        proc = Popen(command, shell=True, stdout=PIPE, stderr=PIPE, stdin=PIPE)
        output, errors = proc.communicate(input=in_text)
    else:
        proc = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
        output, errors = proc.communicate()
    if proc.returncode:
        raise ReturnCode(proc.returncode, o2str(errors), output=o2str(output))
    return o2str(output)


def execute(command, in_text=None):
    """Execute command, returns stdout; prints errors to stderr."""
    dbg(f"execute: `{command}`")
    if in_text is None:
        proc = Popen(command, shell=False, stdout=PIPE, stderr=PIPE)
    else:
        proc = Popen(command, shell=False, stdout=PIPE, stderr=PIPE, stdin=PIPE)
        dbg(f"execute: polling ({in_text})...")
        while proc.poll() is None and proc.stdin is not None:
            dbg("execute: Sending to process stdin")
            proc.stdin.write(in_text)
            dbg("execute: sleeping")
            sleep(0.01)
    output = proc.stdout.read()
    if proc.returncode:
        dbg(f"Received return code {proc.returncode}")
        raise ReturnCode(proc.returncode, proc.stderr.readlines(), output=output)
    return o2str(output)


def execute_with_input(command, in_text):
    """Execute command, passing in_text to stdin if provided."""
    execute(command, in_text)



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2019 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""A wrapper around a Launchpad Personal Package Archive object."""

import os
import re
import sys
import enum

from datetime import datetime
from functools import lru_cache
from itertools import chain
from typing import Iterator, List
from lazr.restfulclient.errors import BadRequest, NotFound, Unauthorized

from .constants import URL_AUTOPKGTEST, DISTRO_UBUNTU_POCKETS
from .io import open_url
from .job import Job, get_waiting, get_running
from .result import get_results
from .source_publication import SourcePublication


class PendingReason(enum.Enum):
    """This describes the reason that leads an operation to hang"""
    BUILD_FAILED = enum.auto()  # Build failed
    BUILD_WAITING = enum.auto()  # Build is still ongoing
    BUILD_PUB_WAITING = enum.auto()  # Build is awaiting publication
    SOURCE_PUB_WAITING = enum.auto()  # Source is awaiting publication
    BUILD_MISSING = enum.auto()  # Build was expected, but missing


class PpaNotFoundError(Exception):
    """Exception indicating a requested PPA could not be found."""

    def __init__(self, ppa_name, owner_name, message=None):
        """Initialize the exception object.

        :param str ppa_name: The name of the missing PPA.
        :param str owner_name: The person or team the PPA belongs to.
        :param str message: An error message.
        """
        self.ppa_name = ppa_name
        self.owner_name = owner_name
        self.message = message

    def __str__(self):
        """Return a human-readable error message.

        :rtype str:
        :return: Error message about the failure.
        """
        if self.message:
            return self.message
        return f"The PPA '{self.ppa_name}' does not exist for person or team '{self.owner_name}'"


class Ppa:
    """Encapsulate data needed to access and conveniently wrap a PPA.

    This object proxies a PPA, allowing lazy initialization and caching
    of data from the remote.
    """

    BUILD_FAILED_FORMAT = (
        "  - {package_name} ({version}) {arch} {state}\n"
        "    + Log Link: {log_link}\n"
        "    + Launchpad Build Page: {build_page}\n"
    )

    def __init__(self, ppa_name, owner_name, ppa_description=None, service=None):
        """Initialize a new Ppa object for a given PPA.

        This creates only the local representation of the PPA, it does
        not cause a new PPA to be created in Launchpad.  For that, see
        PpaGroup.create()

        :param str ppa_name: The name of the PPA within the owning
            person or team namespace.
        :param str owner_name: The name of the person or team the PPA
            belongs to.
        :param str ppa_description: Optional description text for the PPA.
        :param launchpadlib.service service: The Launchpad service object.
        """
        if not ppa_name:
            raise ValueError("undefined ppa_name.")
        if not owner_name:
            raise ValueError("undefined owner_name.")

        self.ppa_name = ppa_name
        self.owner_name = owner_name
        if ppa_description is None:
            self.ppa_description = ''
        else:
            self.ppa_description = ppa_description
        self._service = service

    def __repr__(self) -> str:
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (f'{self.__class__.__name__}('
                f'ppa_name={self.ppa_name!r}, owner_name={self.owner_name!r})')

    def __str__(self) -> str:
        """Return a displayable string identifying the PPA.

        :rtype: str
        :returns: Displayable representation of the PPA.
        """
        return f"{self.owner_name}/{self.name}"

    @property
    @lru_cache
    def archive(self):
        """Retrieve the LP Archive object from the Launchpad service.

        :rtype: archive
        :returns: The Launchpad archive object.
        :raises PpaNotFoundError: Raised if a PPA does not exist in Launchpad.
        """
        if not self._service:
            raise AttributeError("Ppa object not connected to the Launchpad service")
        try:
            owner = self._service.people[self.owner_name]
            return owner.getPPAByName(name=self.ppa_name)
        except NotFound:
            raise PpaNotFoundError(self.ppa_name, self.owner_name)

    @lru_cache
    def exists(self) -> bool:
        """Indicate if the PPA exists in Launchpad."""
        try:
            self.archive
            return True
        except PpaNotFoundError:
            return False

    @property
    @lru_cache
    def address(self):
        """The proper identifier of the PPA.

        :rtype: str
        :returns: The full identification string for the PPA.
        """
        return "ppa:{}/{}".format(self.owner_name, self.ppa_name)

    @property
    def name(self):
        """The name portion of the PPA's address.

        :rtype: str
        :returns: The name of the PPA.
        """
        return self.ppa_name

    @property
    def url(self):
        """The HTTP url for the PPA in Launchpad.

        :rtype: str
        :returns: The url of the PPA.
        """
        return self.archive.web_link

    @property
    def description(self):
        """The description body for the PPA.

        :rtype: str
        :returns: The description body for the PPA.
        """
        return self.ppa_description

    def set_description(self, description):
        """Configure the displayed description for the PPA.

        :rtype: bool
        :returns: True if successfully set description, False on error.
        """
        self.ppa_description = description
        try:
            archive = self.archive
        except PpaNotFoundError as e:
            print(e)
            return False
        archive.description = description
        retval = archive.lp_save()
        print("setting desc to '{}'".format(description))
        print("desc is now '{}'".format(self.archive.description))
        return retval and self.archive.description == description

    @property
    @lru_cache
    def is_private(self) -> bool:
        """Indicates if the PPA is private or public.

        :rtype: bool
        :returns: True if the archive is private, False if public.
        """
        return self.archive.private

    def set_private(self, private: bool):
        """Attempts to configure the PPA as private.

        Note that PPAs can't be changed to private if they ever had any
        sources published, or if the owning person or team is not
        permitted to hold private PPAs.

        :param bool private: Whether the PPA should be private or public.
        """
        if private is None:
            return
        self.archive.private = private
        self.archive.lp_save()

    @property
    @lru_cache
    def publish(self):
        return self.archive.publish

    def set_publish(self, publish: bool):
        if publish is None:
            return
        self.archive.publish = publish
        self.archive.lp_save()

    @property
    @lru_cache
    def architectures(self) -> List[str]:
        """The architectures configured to build packages in the PPA.

        :rtype: List[str]
        :returns: List of architecture names, or None on error.
        """
        try:
            return [proc.name for proc in self.archive.processors]
        except PpaNotFoundError as e:
            sys.stderr.write(e)
            return None

    def set_architectures(self, architectures: List[str]) -> bool:
        """Configure the architectures used to build packages in the PPA.

        Note that some architectures may only be available upon request
        from Launchpad administrators.  ppa.constants.ARCHES_PPA is a
        list of standard architectures that don't require permissions.

        :param List[str] architectures: List of processor architecture names
        :rtype: bool
        :returns: True if architectures could be set, False on error or
            if no architectures were specified.
        """
        if not architectures:
            return False
        base = self._service.API_ROOT_URL.rstrip('/')
        procs = []
        for arch in architectures:
            procs.append(f'{base}/+processors/{arch}')
        try:
            self.archive.setProcessors(processors=procs)
            return True
        except PpaNotFoundError as e:
            sys.stderr.write(e)
            return False

    def set_pocket(self, pocket: str):
        f"""Configure the main archive dependency of the PPA to be against a given pocket.

        :param : str pocket: the targeted pocket (must be one of {DISTRO_UBUNTU_POCKETS})
        """

        if pocket not in DISTRO_UBUNTU_POCKETS:
            raise ValueError(f'{pocket} not in {DISTRO_UBUNTU_POCKETS}')

        main_archive = self.archive.distribution.main_archive
        # Using "multiverse" as a component as it matches the default value on the web UI.
        # See https://git.launchpad.net/launchpad/tree/lib/lp/soyuz/browser/archive.py#n1912
        self.archive.addArchiveDependency(
                component="multiverse",
                dependency=main_archive,
                pocket=pocket.title()
        )
        self.__class__.pocket.fget.cache_clear()

    @property
    @lru_cache
    def pocket(self):
        """The target pocket for the dependency to the main archive."""
        main_archive = self.archive.distribution.main_archive
        for dep in self.archive.dependencies:
            if dep.dependency == main_archive:
                return dep.pocket.lower()
        # If unspecified, the pocket is '$release-updates'
        return 'updates'

    @property
    @lru_cache
    def dependencies(self) -> List[str]:
        """The additional PPAs configured for building packages in this PPA.

        :rtype: List[str]
        :returns: List of PPA addresses
        """
        ppa_addresses = []
        for dep in self.archive.dependencies:
            ppa_dep = dep.dependency
            ppa_addresses.append(ppa_dep.reference)
        return ppa_addresses

    def set_dependencies(self, ppa_addresses: List[str]):
        """Configure the additional PPAs used to build packages in this PPA.

        This removes any existing PPA dependencies and adds the ones
        in the corresponding list.  If any of these new PPAs cannot be
        found, this routine bails out without changing the current set.

        :param List[str] ppa_addresses: Additional PPAs to add
        """
        base = self._service.API_ROOT_URL.rstrip('/')
        new_ppa_deps = []
        for ppa_address in ppa_addresses:
            owner_name, ppa_name = ppa_address_split(ppa_address)
            new_ppa_dep = f'{base}/~{owner_name}/+archive/ubuntu/{ppa_name}'
            new_ppa_deps.append(new_ppa_dep)

        # TODO: Remove all existing dependencies
#        for ppa_dep in self.archive.dependencies:
#            the_ppa.removeArchiveDependency(ppa_dep)

        # TODO: Not sure what to pass here, maybe a string ala 'main'?
        component = None

        # TODO: Allow setting alternate pockets
        # TODO: Maybe for convenience it should be same as what's set for main archive?
        pocket = 'Release'

        for ppa_dep in new_ppa_deps:
            self.archive.addArchiveDependency(
                component=component,
                dependency=ppa_dep,
                pocket=pocket)
        # TODO: Error checking
        #       This can throw ArchiveDependencyError if the ppa_address does not fit the_ppa
        self.__class__.dependencies.fget.cache_clear()

    def get_binaries(
        self, distro=None, series=None, arch=None, created_since_date=None,
        name=None
    ):
        """Retrieve the binary packages available in the PPA.

        :param distribution distro: The Launchpad distribution object.
        :param str series: The distro's codename for the series.
        :param str arch: The hardware architecture.
        :param datetime created_since_date: Only return binaries that
            were created on or after this date.
        :param str name: Only return binaries with this name.
        :rtype: List[binary_package_publishing_history]
        :returns: List of binaries, or None on error
        """
        if distro is None and series is None and arch is None:
            try:
                return chain(
                    self.archive.getPublishedBinaries(
                        created_since_date=created_since_date, status="Pending",
                        binary_name=name),
                    self.archive.getPublishedBinaries(
                        created_since_date=created_since_date, status="Published",
                        binary_name=name))
            except PpaNotFoundError as e:
                print(e)
                return None

        # TODO: Use SourcePublication and retrieve binaries from it
        print("Unimplemented")
        return []

    def get_source_publications(
        self,
        series: 'str | None' = None,
        created_since_date: 'datetime | None' = None,
        name: 'str | None' = None
    ) -> "Iterator(SourcePublication) | None":
        """Retrieve the source packages in the PPA.

        :param str series: The distro codename for the series.
        :param datetime created_since_date: Only return source publications that
            were created on or after this date.
        :param str name: Only return publications for this source package.

        :rtype: Iterator(SourcePublication) | None
        :returns: Collection of source publications, or None on error.
        """
        try:
            # Iterate through the launchpad publications, filtering by
            # any provided criteria, and then return as SourcePackage objects.
            params = {
                'component_name': None,
                'created_since_date': created_since_date,
                'distro_series': None,
                'order_by_date': True,
                'version': None,
            }
            if name:
                params['source_name'] = name
                params['exact_match'] = True
            if series:
                distro = self.archive.distribution
                params['distro_series'] = distro.getSeries(name_or_version=series)
            for lp_publication in chain(
                self.archive.getPublishedSources(status="Pending", **params),
                self.archive.getPublishedSources(status="Published", **params)
            ):
                yield SourcePublication(
                    package=lp_publication.source_package_name,
                    version=lp_publication.source_package_version,
                    series=lp_publication.distro_series.name,
                    archive=self,
                    status=lp_publication.status,
                )
        except PpaNotFoundError as e:
            print(e)
            return None

        return None

    def destroy(self):
        """Delete the PPA.

        :rtype: bool
        :returns: True if PPA was successfully deleted, is in process of
            being deleted, no longer exists, or didn't exist to begin with.
            False if the PPA could not be deleted for some reason and is
            still existing.
        """
        try:
            return self.archive.lp_delete()
        except PpaNotFoundError as e:
            print(e)
            return True
        except BadRequest:
            # Will report 'Archive already deleted' if deleted but not yet gone
            # we can treat this as successfully destroyed
            return True

    def has_packages(self, created_since_date=None, name=None) -> bool:
        """Indicate whether the PPA has any source packages.

        :param created_since_date: Cutoff date for the search, None means no cutoff.
        :param name: Only return source packages with this name.

        :rtype: bool
        :returns: True if PPA contains packages, False if empty or doesn't exit.
        """
        return any(self.archive.getPublishedSources(
            created_since_date=created_since_date,
            source_name=name
        ))

    def pending_publications(
        self,
        created_since_date: 'datetime | None' = None,
        name: 'str | None' = None,
        logging: 'bool' = False
    ) -> 'List[PendingReason]':
        """
        Check for pending publications and returns a list of PendingReason.

        :param datetime created_since_date: Cutoff date for the search, None means no cutoff
        :param str name: Only return pending publications for this source package.

        :rtype: list[PendingReason]
        :returns: A list of PendingReason indicating the status of the
            pending publications. Empty means there are no pending
            publications.
        """
        pending_publication_sources = {}
        required_builds = {}
        pending_publication_builds = {}
        published_builds = {}

        for source_publication in self.get_source_publications(
                created_since_date=created_since_date,
                name=name
        ):
            if not source_publication.date_published:
                pending_publication_sources[source_publication.self_link] = source_publication

            # iterate over the getBuilds result with no status restriction to get build records
            for build in source_publication.getBuilds():
                required_builds[build.self_link] = build

        for binary_publication in self.get_binaries(
                created_since_date=created_since_date,
                name=name
        ):
            # Ignore failed builds
            build = binary_publication.build
            if build.buildstate != "Successfully built":
                continue
            # Skip binaries for obsolete sources
            source_publication = build.current_source_publication
            if source_publication is None:
                continue

            if binary_publication.status == "Pending":
                pending_publication_builds[binary_publication.build_link] = binary_publication
            elif binary_publication.status == "Published":
                published_builds[binary_publication.build_link] = binary_publication

        if not logging:
            os.system('clear')

        retval = []
        num_builds_waiting = (
            len(required_builds) - len(pending_publication_builds) - len(published_builds)
        )
        if num_builds_waiting != 0:
            num_build_failures = 0
            builds_waiting_output = ''
            builds_failed_output = ''
            for build in required_builds.values():
                if build.buildstate == "Successfully built":
                    continue
                elif build.buildstate == "Cancelled build":
                    continue
                elif build.buildstate == "Failed to build":
                    num_build_failures += 1
                    builds_failed_output += self.BUILD_FAILED_FORMAT.format(
                        package_name=build.source_package_name,
                        version=build.source_package_version,
                        arch=build.arch_tag,
                        state=build.buildstate,
                        log_link=build.build_log_url,
                        build_page=build.web_link)
                else:
                    builds_waiting_output += "  - {} ({}) {}: {}\n".format(
                        build.source_package_name,
                        build.source_package_version,
                        build.arch_tag,
                        build.buildstate)
            if num_builds_waiting <= num_build_failures:
                print("* Some builds have failed:")
                print(builds_failed_output)
                retval.append(PendingReason.BUILD_FAILED)
            elif builds_waiting_output != '':
                print("* Still waiting on these builds:")
                print(builds_waiting_output)
                retval.append(PendingReason.BUILD_WAITING)

        if len(pending_publication_builds) != 0:
            num = len(pending_publication_builds)
            print(f"* Still waiting on {num} build publications:")
            for pub in pending_publication_builds.values():
                print("  - {}".format(pub.display_name))
            retval.append(PendingReason.BUILD_PUB_WAITING)
        if len(pending_publication_sources) != 0:
            num = len(pending_publication_sources)
            print(f"* Still waiting on {num} source publications:")
            for pub in pending_publication_sources.values():
                print("  - {}".format(pub.display_name))
            retval.append(PendingReason.SOURCE_PUB_WAITING)
        if ((list(required_builds.keys()).sort() != list(published_builds.keys()).sort())):
            print("* Missing some builds")
            retval.append(PendingReason.BUILD_MISSING)

        if not retval:
            print("Successfully published all builds for all architectures")
        return retval

    def get_autopkgtest_waiting(
            self,
            releases: 'List[str] | None',
            sources: 'List[str] | None' = None
            ) -> Iterator[Job]:
        """Return iterator of queued autopkgtests for this PPA.

        See get_waiting() for details

        :param List[str] releases: The Ubuntu series codename(s), or None.
        :param List[str] sources: Only retrieve results for these
            source packages, or all if blank or None.
        :rtype: Iterator[Job]
        :returns: Currently waiting jobs, if any, or an empty list on error
        """
        response = open_url(f"{URL_AUTOPKGTEST}/queues.json", "waiting autopkgtests")
        if response:
            return get_waiting(response, releases=releases, sources=sources, ppa=str(self))
        return []

    def get_autopkgtest_running(
            self,
            releases: 'List[str] | None',
            sources: 'List[str] | None' = None
            ) -> Iterator[Job]:
        """Return iterator of queued autopkgtests for this PPA.

        See get_running() for details

        :param List[str] releases: The Ubuntu series codename(s), or None.
        :param List[str] packages: Only retrieve results for these
            source packages, or all if blank or None.
        :rtype: Iterator[Job]
        :returns: Currently running jobs, if any, or an empty list on error
        """
        response = open_url(f"{URL_AUTOPKGTEST}/static/running.json", "running autopkgtests")
        if response:
            return get_running(response, releases=releases, sources=sources, ppa=str(self))
        return []

    def get_autopkgtest_results(
            self,
            releases: 'List[str] | None',
            architectures: 'List[str] | None',
            sources: 'List[str] | None' = None
            ) -> Iterator[dict]:
        """Returns iterator of results from autopkgtest runs for this PPA.

        See get_results() for details

        :param list[str] releases: The Ubuntu series codename(s), or None.
        :param list[str] architectures: The hardware architectures.
        :param list[str] sources: Only retrieve results for these
            source packages, or all if blank or None.
        :rtype: Iterator[dict]
        :returns: Autopkgtest results, if any, or an empty list on error
        """
        results = []
        for release in releases:
            base_results_fmt = f"{URL_AUTOPKGTEST}/results/autopkgtest-%s-%s-%s/"
            base_results_url = base_results_fmt % (release, self.owner_name, self.name)
            response = open_url(f"{base_results_url}?format=plain")
            if response:
                trigger_sets = {}
                for result in get_results(
                        response=response,
                        base_url=base_results_url,
                        arches=architectures,
                        sources=sources):
                    trigger = ', '.join([str(r) for r in result.get_triggers()])
                    trigger_sets.setdefault(trigger, [])
                    trigger_sets[trigger].append(result)
                results.append(trigger_sets)
        return results


def ppa_address_split(ppa_address):
    """Parse an address for a PPA into its owner and name components.

    :param str ppa_address: A ppa name or address.
    :rtype: tuple(str, str)
    :returns: The owner name and ppa name as a tuple, or (None, None) on error.
    """
    owner_name = None

    if not ppa_address or len(ppa_address) < 2:
        return (None, None)
    if ppa_address.startswith('ppa:'):
        if '/' not in ppa_address:
            return (None, None)
        rem = ppa_address.split('ppa:', 1)[1]
        owner_name = rem.split('/', 1)[0]
        ppa_name = rem.split('/', 1)[1]
    elif ppa_address.startswith('http'):
        # Only launchpad PPA urls are supported
        m = re.search(
            r'https://launchpad\.net/~([^/]+)/\+archive/ubuntu/([^/]+)(?:/*|/\+[a-z]+)$',
            ppa_address)
        if not m:
            return (None, None)
        owner_name = m.group(1)
        ppa_name = m.group(2)
    elif '/' in ppa_address:
        owner_name = ppa_address.split('/', 1)[0]
        ppa_name = ppa_address.split('/', 1)[1]
    else:
        ppa_name = ppa_address

    if owner_name is not None:
        if len(owner_name) < 1:
            return (None, None)
        owner_name = owner_name.lower()

    if (ppa_name
        and not (any(x.isupper() for x in ppa_name))
        and ppa_name.isascii()
        and '/' not in ppa_name
        and len(ppa_name) > 1):
        return (owner_name, ppa_name)

    return (None, None)


def get_ppa(lp, config):
    """Retrieve the specified PPA from Launchpad.

    :param Lp lp: The Launchpad wrapper object.
    :param dict config: Configuration param:value map.
    :rtype: Ppa
    :returns: Specified PPA as a Ppa object.
    """
    return Ppa(
        ppa_name=config.get('ppa_name', None),
        owner_name=config.get('owner_name', None),
        service=lp)


if __name__ == "__main__":
    import pprint
    import random
    import string
    from .lp import Lp
    from .ppa_group import PpaGroup

    pp = pprint.PrettyPrinter(indent=4)

    print('##########################')
    print('## Ppa class smoke test ##')
    print('##########################')
    print()

    # pylint: disable-next=invalid-name
    rndstr = str(''.join(random.choices(string.ascii_lowercase, k=6)))
    dep_name = f'dependency-ppa-{rndstr}'
    smoketest_ppa_name = f'test-ppa-{rndstr}'

    lp = Lp('smoketest', staging=True)
    ppa_group = PpaGroup(service=lp, name=lp.me.name)

    dep_ppa = ppa_group.create(dep_name, ppa_description=dep_name)
    the_ppa = ppa_group.create(smoketest_ppa_name, ppa_description=smoketest_ppa_name)
    ppa_dependencies = [f'ppa:{lp.me.name}/{dep_name}']

    try:
        the_ppa.set_publish(True)

        if not the_ppa.exists():
            print("Error: PPA does not exist")
            sys.exit(1)
        the_ppa.set_description("This is a testing PPA and can be deleted")
        the_ppa.set_publish(False)
        the_ppa.set_architectures(["amd64", "arm64"])
        the_ppa.set_dependencies(ppa_dependencies)

        print()
        print(f"name:          {the_ppa.name}")
        print(f"address:       {the_ppa.address}")
        print(f"str(ppa):      {the_ppa}")
        print(f"reference:     {the_ppa.archive.reference}")
        print(f"self_link:     {the_ppa.archive.self_link}")
        print(f"web_link:      {the_ppa.archive.web_link}")
        print(f"description:   {the_ppa.description}")
        print(f"has_packages:  {the_ppa.has_packages()}")
        print(f"architectures: {'/'.join(the_ppa.architectures)}")
        print(f"dependencies:  {','.join(the_ppa.dependencies)}")
        print(f"url:           {the_ppa.url}")
        print()

    except BadRequest as e:
        print(f"Error: (BadRequest) {str(e.content.decode('utf-8'))}")
    except Unauthorized as e:
        print(f"Error: (Unauthorized) {e}")

    # pylint: disable-next=invalid-name
    answer = 'x'
    while answer not in ['y', 'n']:
        answer = input('Ready to cleanup (i.e. delete) temporary test PPAs? (y/n) ')
        answer = answer[0].lower()

    if answer == 'y':
        print("  Cleaning up temporary test PPAs...")
        the_ppa.destroy()
        dep_ppa.destroy()
        print("  ...Done")



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2023 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests SourcePackage as an interface to an Apt source package record."""

import os
import sys

import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.source_package import SourcePackage


@pytest.mark.parametrize('pkg_dict, expected_repr, expected_str', [
    ({'package': 'a', 'version': 'b', 'binary': 'c'},
     "SourcePackage(pkg_dict={'package': 'a', 'version': 'b', 'binary': 'c'})",
     "a (b)",
     ),
])
def test_object(pkg_dict, expected_repr, expected_str):
    """Checks that SourcePackage objects can be instantiated."""
    source_package = SourcePackage(pkg_dict)

    assert source_package
    assert repr(source_package) == expected_repr
    assert str(source_package) == expected_str


@pytest.mark.parametrize('pkg_dict, expected_exception', [
    ({}, ValueError),
    ({'package': 'x'}, ValueError),
    ({'version': 'x'}, ValueError),
    ({'binary': 'x'}, ValueError),
    ({'package': None, 'version': None, 'binary': None}, ValueError),
    ({'package': None, 'version': 'x',  'binary': 'x'}, ValueError),
    ({'package': 'x',  'version': None, 'binary': 'x'}, ValueError),
    ({'package': 'x',  'version': 'x',  'binary': None}, ValueError),
])
def test_object_error(pkg_dict, expected_exception):
    """Checks that SourcePackage objects can be instantiated."""
    with pytest.raises(expected_exception):
        source_package = SourcePackage(pkg_dict)
        assert source_package


@pytest.mark.parametrize('pkg_dict, expected_binaries', [
    ({'package': 'x', 'version': 'x', 'binary': 'a, b, c'},
     {'a': '', 'b': '', 'c': ''}),
    ({'package': 'dovecot',
      'version': '1:2.3.19.1+dfsg1-2ubuntu4',
      'binary': 'dovecot-core, dovecot-dev, dovecot-imapd, dovecot-pop3d, dovecot-lmtpd'},
     {'dovecot-core': '',
      'dovecot-dev': '',
      'dovecot-imapd': '',
      'dovecot-pop3d': '',
      'dovecot-lmtpd': '',
      }),
])
def test_provides_binaries(pkg_dict, expected_binaries):
    """Checks that SourcePackage objects parse their Provides field."""
    source_package = SourcePackage(pkg_dict)

    assert source_package.provides_binaries == expected_binaries


@pytest.mark.parametrize('pkg_dict, expected_build_dependencies', [
    ({'package': 'x', 'version': 'x', 'binary': '', 'build-depends': 'a, b, c [x]'},
     {'a': '', 'b': '', 'c': '[x]'}),
])
def test_build_dependencies(pkg_dict, expected_build_dependencies):
    """Checks that SourcePackage objects parse BuildDepends field."""
    source_package = SourcePackage(pkg_dict)

    assert source_package.build_dependencies == expected_build_dependencies



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2022 Authors
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

"""Interprets and analyzes an Ubuntu Apt suite (aka release-pocket)."""

import os.path
from functools import lru_cache
from typing import Dict, List

# pylint: disable = no-name-in-module
import apt_pkg

from .source_package import SourcePackage
from .binary_package import BinaryPackage
from .constants import (
    DISTRO_UBUNTU_COMPONENTS,
    DISTRO_UBUNTU_POCKETS,
    LOCAL_REPOSITORY_PATH,
    LOCAL_REPOSITORY_MIRRORING_DIRECTIONS
)


class Suite:
    """A pocket of a Ubuntu series collecting source and binary package releases.

    Suites are named "<series>-<pocket>", such as "focal-updates" or
    "jammy-proposed".  The same package can have different versions in
    each Suite, but within a Suite each package will have no more than
    one version available at a time.
    """
    def __init__(self, suite_name: str, cache_dir: str):
        """Initialize a new Suite object for a given release pocket.

        :param str series_codename: The textual name of the Ubuntu release.
        :param str pocket: The pocket name ('release', 'proposed', 'backports', etc.)
        :param str cache_dir: The path to the given suite in the local Apt mirror.
        """
        if not suite_name:
            raise ValueError('undefined suite_name.')
        if not cache_dir:
            raise ValueError('undefined cache_dir.')
        if not os.path.exists(cache_dir):
            raise FileNotFoundError(f"could not find cache dir '{cache_dir}'")

        self._suite_name = suite_name
        self._cache_dir = cache_dir
        self._provides_table = None
        self._rdepends_table = None

    def __repr__(self) -> str:
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (
            f'{self.__class__.__name__}('
            f'suite_name={self._suite_name!r}, '
            f'cache_dir={self._cache_dir!r})'
        )

    def __str__(self) -> str:
        """Return a human-readable textual description of the Suite.

        :rtype: str
        :returns: Human-readable string.
        """
        return f'{self._suite_name}'

    def _rebuild_lookup_tables(self) -> bool:
        """Regenerate the provides and rdepends lookup tables.

        Some packages have build dependence that can be satisfied by one
        of several packages.  For example, a package may require either
        awk or mawk to build.  In these cases, the package will be
        registered in the table as an rdepend for BOTH awk and mawk.

        :rtype: bool
        :returns: True if tables were rebuilt, False otherwise"""
        self._provides_table = {}
        self._rdepends_table = {}
        for source_name, source in self.sources.items():
            for build_dep_binary_names in source.build_dependencies.keys():
                # This needs to deal with two different kinds of keys.
                # Basic dependencies are just simple str's, while alternate
                # dependencies are modeled as tuples.
                #
                # So, convert simple str's into single-element lists, so
                # both cases can be handled via iteration in a for loop.
                if isinstance(build_dep_binary_names, str):
                    build_dep_binary_names = [build_dep_binary_names]
                for build_dep_binary_name in build_dep_binary_names:
                    self._rdepends_table.setdefault(build_dep_binary_name, [])
                    self._rdepends_table[build_dep_binary_name].append(source)

            for provided_binary_name in source.provides_binaries.keys():
                self._provides_table[provided_binary_name] = source
        return self._provides_table and self._rdepends_table

    @property
    @lru_cache
    def info(self) -> Dict[str, str]:
        """The parsed Apt Release file for the suite as a dict.

        :rtype: dict[str, str]
        """
        info = None
        with apt_pkg.TagFile(f'{self._cache_dir}/Release') as tagfile:
            info = next(tagfile)
        if not info:
            raise ValueError(f'Could not load {self._cache_dir}/Release')
        return info

    @property
    def name(self) -> str:
        """The name of the suite as recorded in the apt database.

        :rtype: str
        """
        suite_name = self.info.get('Suite')
        if not suite_name:
            raise ValueError('Could not get suite name from info dict.')
        return suite_name

    @property
    def series_codename(self) -> str:
        """The textual name of the Ubuntu release for this suite.

        :rtype: str
        """
        return self.name.split('-')[0]

    @property
    def pocket(self) -> str:
        """The category of the archive (release, proposed, security, et al).

        :rtype: str
        """
        if '-' not in self.name:
            return 'release'
        pocket = self.name.split('-')[1]
        if pocket not in DISTRO_UBUNTU_POCKETS:
            raise RuntimeError(f'Unrecognized pocket "{pocket}"')
        return pocket

    @property
    def architectures(self) -> List[str]:
        """The list of CPU hardware types supported by this suite.

        :rtype: list[str]
        """
        architectures = self.info.get('Architectures').split()
        if not architectures:
            raise RuntimeError('Could not load architectures from info')
        return architectures

    @property
    def components(self) -> List[str]:
        """The sections of the archive provided in this suite.

        Components may include main, universe, etc.

        :rtype: list[str]
        """
        components = [
            component
            for component in os.listdir(self._cache_dir)
            if os.path.isdir(os.path.join(self._cache_dir, component))
            and component in DISTRO_UBUNTU_COMPONENTS
        ]
        if not components:
            raise RuntimeError(f'Could not load components from {self._cache_dir}')
        return components

    @property
    @lru_cache
    def sources(self) -> Dict[str, SourcePackage]:
        """The collection of source packages included in this suite.

        All source packages in all components are returned as
        SourcePackage objects.

        :rtype: dict[str, SourcePackage]
        """
        sources = None
        for sources_file in ['Sources.xz', 'Sources.gz']:
            for comp in self.components:
                source_packages_dir = f'{self._cache_dir}/{comp}/source'
                try:
                    with apt_pkg.TagFile(f'{source_packages_dir}/{sources_file}') as pkgs:
                        if sources is None:
                            sources = {}
                        for pkg in pkgs:
                            name = pkg['Package']
                            sources[name] = SourcePackage(dict(pkg))
                except apt_pkg.Error:
                    pass
            if sources is not None:
                return sources
        raise RuntimeError(f'Could not load {source_packages_dir}/Sources.[xz|gz]')

    @property
    @lru_cache
    def binaries(self) -> Dict[str, BinaryPackage]:
        """The collection of binary Deb packages included in this suite.

        All binary packages in all components are returned as
        BinaryPackage objects.

        :rtype: dict[str, BinaryPackage]
        """
        binaries = None
        for packages_file in ["Packages.xz", "Packages.gz"]:
            for comp in self.components:
                for arch in self.architectures:
                    binary_packages_dir = f'{self._cache_dir}/{comp}/binary-{arch}'
                    try:
                        with apt_pkg.TagFile(f'{binary_packages_dir}/{packages_file}') as pkgs:
                            if binaries is None:
                                binaries = {}
                            for pkg in pkgs:
                                name = f'{pkg["Package"]}:{arch}'
                                binaries[name] = BinaryPackage(pkg)
                    except apt_pkg.Error:
                        pass
            if binaries is not None:
                return binaries
        raise ValueError(f'Could not load {binary_packages_dir}/Packages.[xz|gz]')

    def dependent_packages(self, source_package: SourcePackage) -> Dict[str, SourcePackage]:
        """Return relevant packages to run autotests against for a given source package.

        Calculates the collection of reverse dependencies for a given
        source package that would be appropriate to re-run autopkgtests
        on, using the given @param source_package's name as a trigger.

        For leaf packages (that nothing else depends on as a build
        requirement), this routine returns an empty dict.

        For packages that can serve as an alternative dependency of some
        packages, this will include all such packages as if they were
        hard dependencies.  For example, when examining postgresql-12, this
        would include all packages dependent on any database.

        :param str source_package_name: The archive name of the source package.
        :rtype: dict[str, SourcePackage]
        :returns: Collection of source packages, keyed by name.
        """
        # Build the lookup table for provides and rdepends
        if not self._rdepends_table:
            if not self._rebuild_lookup_tables():
                raise RuntimeError("Could not regenerate provides/rdepends lookup tables")

        dependencies = {}

        # Get source packages that depend on things we supply
        for binary_package_name in source_package.provides_binaries.keys():
            rdeps = self._rdepends_table.get(binary_package_name)
            if rdeps:
                for rdep_source in rdeps:
                    dependencies[rdep_source.name] = rdep_source

        return dependencies


if __name__ == '__main__':
    # pylint: disable=invalid-name
    import sys
    from pprint import PrettyPrinter
    pp = PrettyPrinter(indent=4)

    from .repository import Repository
    from .debug import error

    print('############################')
    print('## Suite class smoke test ##')
    print('############################')
    print()

    local_dists_path = os.path.join(LOCAL_REPOSITORY_PATH, 'dists')
    if not os.path.exists(local_dists_path):
        error(f'Missing checkout for suite smoketest\n{LOCAL_REPOSITORY_MIRRORING_DIRECTIONS}')
        sys.exit(1)

    repository = Repository(cache_dir=local_dists_path)
    for suite in repository.suites.values():
        print(suite)
        print(f'  series:         {suite.series_codename}')
        print(f'  pocket:         {suite.pocket}')
        print(f'  components:     {", ".join(suite.components)}')
        print(f'  architectures:  {", ".join(suite.architectures)}')

        num_sources = len(suite.sources)
        ellipses_shown = False
        print(f'  sources:        ({num_sources} items)')
        for i, suite_source in enumerate(suite.sources):
            if i < 3 or i >= num_sources - 3:
                print(f'    {i} {suite_source}')
            elif not ellipses_shown:
                print('    [...]')
                ellipses_shown = True

        num_binaries = len(suite.binaries)
        ellipses_shown = False
        print(f'  binaries:       ({num_binaries} items)')
        for i, binary in enumerate(suite.binaries):
            if i < 3 or i >= num_binaries - 3:
                print(f'    {i} {binary}')
            elif not ellipses_shown:
                print('    [...]')
                ellipses_shown = True



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2023 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

# pylint: disable=protected-access,line-too-long

"""Tests the Suite class as an interface to Apt suite records."""

import os
import sys

import lzma as xz
import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')))

from ppa.suite import Suite
from ppa.source_package import SourcePackage
from ppa.binary_package import BinaryPackage


@pytest.mark.parametrize('suite_name, cache_dir, expected_repr, expected_str', [
    ('x', '/tmp', "Suite(suite_name='x', cache_dir='/tmp')", "x"),
    ('a-1', '/tmp', "Suite(suite_name='a-1', cache_dir='/tmp')", "a-1"),
    ('b-2', '/tmp', "Suite(suite_name='b-2', cache_dir='/tmp')", "b-2"),
])
def test_object(suite_name, cache_dir, expected_repr, expected_str):
    """Checks that Suite objects can be instantiated."""
    suite = Suite(suite_name, cache_dir)

    assert suite
    assert repr(suite) == expected_repr
    assert str(suite) == expected_str


@pytest.mark.parametrize('suite_name, cache_dir, expected_exception', [
    ('x', '', ValueError),
    ('', 'x', ValueError),
    ('x', 'x', FileNotFoundError),
    ('', '', ValueError),
    ('a-1', None, ValueError),
    (None, 'x', ValueError),
    (None, None, ValueError),
])
def test_object_error(suite_name, cache_dir, expected_exception):
    """Checks that Suite objects handle invalid input properly."""
    with pytest.raises(expected_exception):
        suite = Suite(suite_name, cache_dir)
        assert suite


@pytest.mark.parametrize('release_contents, expected_info', [
    ('a: 1', {'a': '1'}),
    ("""
Origin: Ubuntu
Label: Ubuntu
Suite: lunar
Version: 23.04
Codename: lunar
Date: Tue, 28 Feb 2023 19:49:32 UTC
Architectures: amd64 arm64 armhf i386 ppc64el riscv64 s390x
Components: main restricted universe multiverse
Description: Ubuntu Lunar 23.04
    """, {
        'Suite': 'lunar',
        'Codename': 'lunar',
        'Architectures': 'amd64 arm64 armhf i386 ppc64el riscv64 s390x',
        'Components': 'main restricted universe multiverse',
    }), ("""
Origin: Ubuntu
Label: Ubuntu
Suite: lunar-proposed
Version: 23.04
Codename: lunar
Date: Tue, 28 Feb 2023 19:50:27 UTC
Architectures: amd64 arm64 armhf i386 ppc64el riscv64 s390x
Components: main restricted universe multiverse
Description: Ubuntu Lunar Proposed
NotAutomatic: yes
ButAutomaticUpgrades: yes
MD5Sum:
 7de6b7c0ed6b4bfb662e07fbc449dfdd        148112816 Contents-amd64
    """, {
        'Suite': 'lunar-proposed',
        'Codename': 'lunar',
        'NotAutomatic': 'yes'
    }),
])
def test_info(tmp_path, release_contents, expected_info):
    """Checks the parsing of info loaded from the Release file."""
    # Create Release file using release_contents in synthetic tree
    suite_dir = tmp_path / 'x'
    suite_dir.mkdir()
    release_file = suite_dir / 'Release'
    release_file.write_text(release_contents)

    suite = Suite(suite_dir.name, suite_dir)

    # Verify the expected items are present in the suite's info dict
    for key, value in expected_info.items():
        assert key in suite.info.keys()
        assert suite.info[key] == value


@pytest.mark.parametrize('info, expected_series_codename', [
    ({'Suite': 'x'}, 'x'),
    ({'Suite': 'x-y'}, 'x'),
    ({'Suite': 'lunar'}, 'lunar'),
    ({'Suite': 'lunar-proposed'}, 'lunar'),
    ({'Suite': 'focal-security'}, 'focal'),
])
def test_series_codename(monkeypatch, info, expected_series_codename):
    """Checks the codename is extracted properly from the suite name."""
    suite = Suite('x', '/tmp')

    # Substitute in our fake test info in place of Suite's info() routine
    monkeypatch.setattr(Suite, "info", info)

    assert suite.series_codename == expected_series_codename


@pytest.mark.parametrize('info, expected_pocket', [
    ({'Suite': 'x'}, 'release'),
    ({'Suite': 'x-backports'}, 'backports'),
    ({'Suite': 'lunar'}, 'release'),
    ({'Suite': 'lunar-proposed'}, 'proposed'),
    ({'Suite': 'focal-security'}, 'security'),
])
def test_pocket(monkeypatch, info, expected_pocket):
    """Checks the pocket is extracted properly from the suite name."""
    suite = Suite('x', '/tmp')

    # Substitute in our fake test info in place of Suite's info() routine
    monkeypatch.setattr(Suite, "info", info)

    assert suite.pocket == expected_pocket


@pytest.mark.parametrize('info, expected_architectures', [
    ({'Architectures': 'x y z'}, ['x', 'y', 'z']),
    ({'Architectures': 'x y z'}, ['x', 'y', 'z']),
    ({'Architectures': 'amd64 arm64 armhf i386 ppc64el riscv64 s390x'},
     ['amd64', 'arm64', 'armhf', 'i386', 'ppc64el', 'riscv64', 's390x']),
])
def test_architectures(monkeypatch, info, expected_architectures):
    """Checks that the architecture list is parsed from the info dict."""
    suite = Suite('x', '/tmp')

    # Substitute in our fake test info in place of Suite's info() routine
    monkeypatch.setattr(Suite, "info", info)

    assert sorted(suite.architectures) == sorted(expected_architectures)


@pytest.mark.parametrize('suite_name, component_paths, expected_components', [
    ('a-1', ['a-1/main', 'a-1/universe', 'a-1/multiverse'], ['main', 'universe', 'multiverse']),
    ('a-1', ['a-1/main', 'b-1/universe', 'c-1/multiverse'], ['main']),
    ('a-1', ['a-1/main', 'a-1/main/y', 'c-1/multiverse/x'], ['main']),
    ('x', ['x/main', 'x/restricted', 'x/universe', 'x/multiverse'],
     ['main', 'restricted', 'universe', 'multiverse']),
])
def test_components(tmp_path, suite_name, component_paths, expected_components):
    """Checks that the components are read from the Apt directory tree.

    The repository could have multiple suites (b-1, c-1, ...)
    so we specify that we're just looking for the components in
    @param suite_name.
    """
    # Stub in suite's directory structure with component subdirs
    for component_path in component_paths:
        component_dir = tmp_path / component_path
        component_dir.mkdir(parents=True)

    suite = Suite(suite_name, tmp_path / suite_name)

    assert sorted(suite.components) == sorted(expected_components)


@pytest.mark.parametrize('sources_contents, expected_sources', [
    ('Package: a\nVersion: b\nBinary: c',
     {'a': SourcePackage({'Package': 'a', 'Version': 'b', 'Binary': 'c'})}),
    ("""
Package: aalib
Version:
Binary:

Package: abseil
Version:
Binary:

Package: accountsservice
Version:
Binary:

Package: acct
Version:
Binary:
    """, {
        'aalib': SourcePackage({'Package': 'aalib', 'Version': '', 'Binary': ''}),
        'abseil': SourcePackage({'Package': 'abseil', 'Version': '', 'Binary': ''}),
        'accountsservice': SourcePackage({'Package': 'accountsservice', 'Version': '', 'Binary': ''}),
        'acct': SourcePackage({'Package': 'acct', 'Version': '', 'Binary': ''}),
    }),
    ("""
Package: libsigc++-2.0
Format: 3.0 (quilt)
Binary: libsigc++-2.0-0v5, libsigc++-2.0-dev, libsigc++-2.0-doc
Architecture: any all
Version: 2.12.0-1
Priority: optional
Section: devel
Maintainer: Debian GNOME Maintainers <pkg-gnome-maintainers@lists.alioth.debian.org>
Uploaders: Jeremy Bicha <jbicha@ubuntu.com>, Michael Biebl <biebl@debian.org>
Standards-Version: 4.6.1
Build-Depends: debhelper-compat (= 13), dh-sequence-gnome, docbook-xml, docbook-xsl, doxygen, graphviz, libxml2-utils <!nocheck>, meson (>= 0.50.0), mm-common (>= 1.0.0), python3-distutils, xsltproc
Homepage: https://libsigcplusplus.github.io/libsigcplusplus/
Vcs-Browser: https://salsa.debian.org/gnome-team/libsigcplusplus
Vcs-Git: https://salsa.debian.org/gnome-team/libsigcplusplus.git
Directory: pool/main/libs/libsigc++-2.0
Package-List:
 libsigc++-2.0-0v5 deb libs optional arch=any
 libsigc++-2.0-dev deb libdevel optional arch=any
 libsigc++-2.0-doc deb doc optional arch=all
Files:
 23feb2cc5036384f94a3882c760a7eb4 2336 libsigc++-2.0_2.12.0-1.dsc
 8685af8355138b1c48a6cd032e395303 163724 libsigc++-2.0_2.12.0.orig.tar.xz
 d60ca8c15750319f52d3b7eaeb6d99e1 10800 libsigc++-2.0_2.12.0-1.debian.tar.xz
Checksums-Sha1:
 81840b1d39dc48350de207566c86b9f1ea1e22d2 2336 libsigc++-2.0_2.12.0-1.dsc
 f66e696482c4ff87968823ed17b294c159712824 163724 libsigc++-2.0_2.12.0.orig.tar.xz
 a699c88f7c91157af4c5cdd0f4d0ddebeea9092e 10800 libsigc++-2.0_2.12.0-1.debian.tar.xz
    """,  # noqa: E501
        {
            'libsigc++-2.0': SourcePackage({
                'Package': 'libsigc++-2.0',
                'Version': '2.12.0-1',
                'Binary': 'libsigc++-2.0-0v5, libsigc++-2.0-dev, libsigc++-2.0-doc',
            }),
        }),
])
def test_sources(tmp_path, sources_contents, expected_sources):
    """Checks that the source packages are read from the Apt record.

    We don't care about the SourcePackage object itself (the value in
    @param expected_sources is just a placeholder), but need to ensure
    the Sources.xz file is read and the expected list of packages
    parsed out of it.
    """
    # Create Sources.xz file using sources_contents in synthetic tree
    suite_dir = tmp_path / 'x'
    suite_dir.mkdir()
    component_dir = suite_dir / 'main'
    component_dir.mkdir()
    arch_dir = component_dir / 'source'
    arch_dir.mkdir()
    sources_file = arch_dir / 'Sources.xz'
    sources_file.write_bytes(xz.compress(str.encode(sources_contents)))

    # Create the suite to wrapper our path and access Sources.xz
    suite = Suite(suite_dir.name, suite_dir)

    assert sorted(suite.sources) == sorted(expected_sources)


@pytest.mark.parametrize('architectures, packages_contents, expected_binaries', [
    (['x'], 'Package: a\nVersion:\nArchitecture: x',
     {'a:x': BinaryPackage({'Package': 'a', 'Version': '', 'Architecture': ''})}),
    (['amd64'], """
Package: aalib
Version:
Architecture:

Package: abseil
Version:
Architecture:

Package: accountsservice
Version:
Architecture:

Package: acct
Version:
Architecture:
    """,
     {
        'aalib:amd64': BinaryPackage({'Package': 'aalib', 'Version': '', 'Architecture': ''}),
        'abseil:amd64': BinaryPackage({'Package': 'abseil', 'Version': '', 'Architecture': ''}),
        'accountsservice:amd64': BinaryPackage({'Package': 'accountsservice', 'Version': '', 'Architecture': ''}),
        'acct:amd64': BinaryPackage({'Package': 'acct', 'Version': '', 'Architecture': ''}),
     }),
    (['amd64'], """
Package: accountsservice
Architecture: amd64
Version: 22.08.8-1ubuntu4
Priority: optional
Section: gnome
Origin: Ubuntu
Maintainer: Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>
Original-Maintainer: Debian freedesktop.org maintainers <pkg-freedesktop-maintainers@lists.alioth.debian.org>
Bugs: https://bugs.launchpad.net/ubuntu/+filebug
Installed-Size: 504
Depends: dbus (>= 1.9.18), libaccountsservice0 (= 22.08.8-1ubuntu4), libc6 (>= 2.34), libglib2.0-0 (>= 2.63.5), libpolkit-gobject-1-0 (>= 0.99)
Recommends: default-logind | logind
Suggests: gnome-control-center
Filename: pool/main/a/accountsservice/accountsservice_22.08.8-1ubuntu4_amd64.deb
Size: 68364
MD5sum: a10447714f0ce3c5f607b7b27c0f9299
SHA1: 252482d8935d16b7fc5ced0c88819eeb7cad6c65
SHA256: 4e341a8e288d8f8f9ace6cf90a1dcc3211f06751d9bec3a68a6c539b0c711282
SHA512: eb91e21b4dfe38e9768e8ca50f30f94298d86f07e78525ab17df12eb3071ea21cfbdc2ade3e48bd4e22790aa6ce3406bb10fbd17d8d668f84de1c8adeee249cb
Homepage: https://www.freedesktop.org/wiki/Software/AccountsService/
Description: query and manipulate user account information
Task: ubuntu-desktop-minimal, ubuntu-desktop, ubuntu-desktop-raspi, ubuntu-wsl, kubuntu-desktop, xubuntu-minimal, xubuntu-desktop, lubuntu-desktop, ubuntustudio-desktop-core, ubuntustudio-desktop, ubuntukylin-desktop, ubuntu-mate-core, ubuntu-mate-desktop, ubuntu-budgie-desktop, ubuntu-budgie-desktop-raspi, ubuntu-unity-desktop, edubuntu-desktop-minimal, edubuntu-desktop, edubuntu-desktop-raspi, edubuntu-wsl
Description-md5: 8aeed0a03c7cd494f0c4b8d977483d7e
    """,
     {  # noqa: E501
       'accountsservice:amd64': BinaryPackage({
           'Package': 'accountsservice',
           'Version': '',
           'Architecture': 'amd64'
        })
     }),
    (
        ['amd64', 'arm64', 'armhf', 'i386', 'ppc64el', 'riscv64', 's390x'],
        'Package: libsigc++-2.0\nVersion:\nArchitecture: amd64 arm64 armhf i386 ppc64el riscv64 s390x\n',
        {
            'libsigc++-2.0:amd64': BinaryPackage({
                'Package': 'libsigc++-2.0', 'Version': '', 'Architecture': 'amd64'}),
            'libsigc++-2.0:arm64': BinaryPackage({
                'Package': 'libsigc++-2.0', 'Version': '', 'Architecture': 'arm64'}),
            'libsigc++-2.0:armhf': BinaryPackage({
                'Package': 'libsigc++-2.0', 'Version': '', 'Architecture': 'armhf'}),
            'libsigc++-2.0:i386': BinaryPackage({
                'Package': 'libsigc++-2.0', 'Version': '', 'Architecture': 'i386'}),
            'libsigc++-2.0:ppc64el': BinaryPackage({
                'Package': 'libsigc++-2.0', 'Version': '', 'Architecture': 'ppc64el'}),
            'libsigc++-2.0:riscv64': BinaryPackage({
                'Package': 'libsigc++-2.0', 'Version': '', 'Architecture': 'riscv64'}),
            'libsigc++-2.0:s390x': BinaryPackage({
                'Package': 'libsigc++-2.0', 'Version': '', 'Architecture': 's390x'}),
        },
    ),
])
def test_binaries(tmp_path, architectures, packages_contents, expected_binaries):
    """Checks that the binary packages are read from the Apt record.

    We don't care about the BinaryPackage (the value) itself, just that
    the package name and arch are registered correctly, and that typical
    Packages.xz files are processed as intended.
    """
    suite_dir = tmp_path / 'x'
    suite_dir.mkdir()
    release_file = suite_dir / 'Release'
    release_file.write_text(f'Architectures: {" ".join(architectures)}')
    component_dir = suite_dir / 'main'
    component_dir.mkdir()
    for arch in architectures:
        arch_dir = component_dir / f'binary-{arch}'
        arch_dir.mkdir()
        packages_file = arch_dir / 'Packages.xz'
        packages_file.write_bytes(xz.compress(str.encode(packages_contents)))

    # Create the suite to wrapper our path and access Packages.xz
    suite = Suite(suite_dir.name, suite_dir)
    assert sorted(suite.binaries.keys()) == sorted(expected_binaries.keys())


@pytest.mark.parametrize('sources, expected_rdepends, expected_provides', [
    ({}, [], []),
    (
        {'a': SourcePackage({'Package': 'a', 'Version': 'x', 'Build-Depends': 'a', 'Binary': 'x'})},
        ['a'],
        ['x']
    ),
    (
        {'a': SourcePackage({'Package': 'a', 'Version': 'x', 'Build-Depends': 'a, b, c', 'Binary': 'x'})},
        ['a', 'b', 'c'],
        ['x']
    ),
    (
        {
            'a': SourcePackage({'Package': 'a', 'Version': 'x', 'Binary': 'a1'}),
            'b': SourcePackage({'Package': 'b', 'Version': 'x', 'Build-Depends': 'a1', 'Binary': 'b1'}),
            'c': SourcePackage({'Package': 'c', 'Version': 'x', 'Build-Depends': 'a1, b1', 'Binary': 'c1'}),
        },
        ['a1', 'b1'],
        ['a1', 'b1', 'c1']
    ),
    (
        {'a': SourcePackage({'Package': 'a', 'Version': 'x', 'Binary': 'a'})},
        [],
        ['a']),
    (
        {'a': SourcePackage({'Package': 'a', 'Version': 'x', 'Binary': 'a, b, c'})},
        [],
        ['a', 'b', 'c']
    ),
    (
        {
            'dovecot': SourcePackage(
                {
                    'Package': 'dovecot',
                    'Version': '1:2.3.19.1+dfsg1-2ubuntu4',
                    'Binary': 'dovecot-core, dovecot-dev, dovecot-imapd, dovecot-pop3d, dovecot-lmtpd, dovecot-managesieved, dovecot-pgsql, dovecot-mysql, dovecot-sqlite, dovecot-ldap, dovecot-gssapi, dovecot-sieve, dovecot-solr, dovecot-lucene, dovecot-submissiond, dovecot-auth-lua',  # noqa: E501
                    'Architecture': 'any',
                    'Build-Depends': 'debhelper-compat (= 13), default-libmysqlclient-dev, krb5-multidev, libapparmor-dev [linux-any], libbz2-dev, libcap-dev [linux-any], libclucene-dev, libdb-dev, libexpat-dev, libexttextcat-dev, libicu-dev, libldap2-dev, liblua5.3-dev, liblz4-dev, liblzma-dev, libpam0g-dev, libpq-dev, libsasl2-dev, libsodium-dev, libsqlite3-dev, libssl-dev, libstemmer-dev, libsystemd-dev [linux-any], libunwind-dev [amd64 arm64 armel armhf hppa i386 ia64 mips mips64 mips64el mipsel powerpc powerpcspe ppc64 ppc64el sh4], libwrap0-dev, libzstd-dev, lsb-release, pkg-config, zlib1g-dev',  # noqa: E501
                    'Testsuite-Triggers': 'lsb-release, python3, systemd-sysv',
                }
            )
        },
        [
            'debhelper-compat',
            'default-libmysqlclient-dev',
            'krb5-multidev',
            'libapparmor-dev',
            'libbz2-dev',
            'libcap-dev',
            'libclucene-dev',
            'libdb-dev',
            'libexpat-dev',
            'libexttextcat-dev',
            'libicu-dev',
            'libldap2-dev',
            'liblua5.3-dev',
            'liblz4-dev',
            'liblzma-dev',
            'libpam0g-dev',
            'libpq-dev',
            'libsasl2-dev',
            'libsodium-dev',
            'libsqlite3-dev',
            'libssl-dev',
            'libstemmer-dev',
            'libsystemd-dev',
            'libunwind-dev',
            'libwrap0-dev',
            'libzstd-dev',
            'lsb-release',
            'pkg-config',
            'zlib1g-dev',
        ],
        [
            'dovecot-core',
            'dovecot-dev',
            'dovecot-imapd',
            'dovecot-pop3d',
            'dovecot-lmtpd',
            'dovecot-managesieved',
            'dovecot-pgsql',
            'dovecot-mysql',
            'dovecot-sqlite',
            'dovecot-ldap',
            'dovecot-gssapi',
            'dovecot-sieve',
            'dovecot-solr',
            'dovecot-lucene',
            'dovecot-submissiond',
            'dovecot-auth-lua',
        ]
    ),
])
def test_rebuild_tables(monkeypatch, sources, expected_rdepends, expected_provides):
    """Checks generation of the internal lookup tables for provides and rdepends."""
    monkeypatch.setattr(Suite, "sources", sources)
    # Verify provides and rdepends table are as expected
    suite = Suite('x', '/tmp')
    suite._rebuild_lookup_tables()

    assert sorted(suite._rdepends_table.keys()) == sorted(expected_rdepends)
    assert sorted(suite._provides_table.keys()) == sorted(expected_provides)


def test_rebuild_tables_mapping(monkeypatch):
    """Checks the mapping of rdepends to provides in the generated tables.

    The two lookup tables are essential to the rdepends test functionality
    since they define the mappings between various source package provides
    and depends.  This test builds a synthetic collection of source packages,
    generates the tables, and then verifies the tables can be used to lookup
    the appropriate related packages.

    For purposes of this test, we assume each source package provides
    binaries of the same name appended with either '1' or '2'.

    Also, note that the packages are set up with a circular dependency
    (a depends on c, but c depends on a).  This is an unhealthy
    situation for an archive to be in, but it certainly does happen in
    the wild.  We're just setting it up that way for convenience since
    we can then assume all provided binaries will be required by
    something in the archive.
    """
    sources = {
        'a': SourcePackage({'Package': 'a', 'Version': 'x', 'Build-Depends': 'c1', 'Binary': 'a1'}),
        'b': SourcePackage({'Package': 'b', 'Version': 'x', 'Build-Depends': 'a1, c2', 'Binary': 'b1'}),
        'c': SourcePackage({'Package': 'c', 'Version': 'x', 'Build-Depends': 'a1, b1', 'Binary': 'c1, c2'}),
    }
    monkeypatch.setattr(Suite, "sources", sources)
    suite = Suite('x', '/tmp')
    suite._rebuild_lookup_tables()

    # Check the integrity of the lookup tables for the sources we gave it
    for source in sources.values():
        # Verify our dependency is satisfied by a SourcePackage in the collection
        for dependency in source.build_dependencies:
            assert dependency in suite._provides_table
            package = suite._provides_table[dependency]
            assert isinstance(package, SourcePackage)
            assert dependency in [f"{package.name}1", f"{package.name}2"]

        # Verify SourcePackages that depend on us can be located
        for binary in source.provides_binaries:
            assert binary in suite._rdepends_table
            for package in suite._rdepends_table[binary]:
                assert isinstance(package, SourcePackage)
                assert binary in package.build_dependencies


@pytest.mark.parametrize('sources_dict, source_package_name, expected_packages', [
    pytest.param(
        [{'Package': 'a', 'Version': 'x', 'Binary': 'a1'}], 'a', [],
        marks=pytest.mark.xfail(reason='Does not handle undefined build-depends'),
    ),
    pytest.param(
        [{'Package': 'a', 'Version': 'x', 'Build-Depends': None, 'Binary': 'a1'}], 'a', [],
        marks=pytest.mark.xfail(reason='Does not handle undefined build-depends'),
    ),
    pytest.param(
        [{'Package': 'a', 'Version': 'x', 'Build-Depends': '', 'Binary': 'a1'}], 'a', [],
        marks=pytest.mark.xfail(reason='Does not handle undefined build-depends'),
    ),

    (
        [
            {'Package': 'a', 'Version': 'x', 'Build-Depends': 'b1', 'Binary': 'a1'},
            {'Package': 'b', 'Version': 'x', 'Build-Depends': 'a1', 'Binary': 'b1'},
            {'Package': 'c', 'Version': 'x', 'Build-Depends': 'b1', 'Binary': 'c1'},
        ],
        'a',
        ['b'],
    ),

    (
        [
            {'Package': 'a', 'Version': 'x', 'Build-Depends': 'd1', 'Binary': 'a1'},
            {'Package': 'b', 'Version': 'x', 'Build-Depends': 'a1', 'Binary': 'b1'},
            {'Package': 'c', 'Version': 'x', 'Build-Depends': 'b1', 'Binary': 'c1, c2'},
            {'Package': 'd', 'Version': 'x', 'Build-Depends': 'a1, b1, c2', 'Binary': 'd1'},
        ],
        'd',
        ['a'],
    ),
])
def test_dependent_packages(monkeypatch, sources_dict, source_package_name, expected_packages):
    '''Checks that dependent_packages() returns the right packages to test.

    This member function is the main API for looking up what packages
    should have autopkgtests run, triggered against our desired package.
    '''
    sources = {pkg['Package']: SourcePackage(pkg) for pkg in sources_dict}

    monkeypatch.setattr(
        Suite,
        "sources",
        sources
    )
    suite = Suite('x', '/tmp')
    source_package = sources[source_package_name]

    assert sorted(suite.dependent_packages(source_package)) == expected_packages



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2019 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

import sys
import pprint
import textwrap

DEBUGGING = False


def dbg(msg, wrap=0, prefix=None, indent=''):
    """Print information if debugging is enabled."""
    if DEBUGGING:
        if type(msg) is str:
            if wrap == 0 and indent != '':
                wrap = 72
            if wrap > 0:
                if prefix is None and len(indent) > 0:
                    prefix = indent
                msg = textwrap.fill(
                    msg,
                    width=wrap,
                    initial_indent=prefix,
                    subsequent_indent=indent)
            sys.stderr.write(f"{msg}\n")
        else:
            pprint.pprint(msg)


def warn(msg):
    """Print warning message to stderr."""
    sys.stderr.write(f"Warning: {msg}\n")


def error(msg):
    """Print error message to stderr."""
    sys.stderr.write(f"Error: {msg}\n")



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2023 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests the BinaryPackage class as an interface to an Apt binary package record."""

import os
import sys

import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.binary_package import BinaryPackage


@pytest.mark.parametrize('pkg_dict, expected_repr, expected_str', [
    ({'package': 'a', 'architecture': 'b', 'version': '1'},
     "BinaryPackage(pkg_dict={'package': 'a', 'architecture': 'b', 'version': '1'})",
     'a (1) [b]'),
])
def test_object(pkg_dict, expected_repr, expected_str):
    """Checks that BinaryPackage objects can be instantiated."""
    obj = BinaryPackage(pkg_dict)

    assert obj
    assert repr(obj) == expected_repr
    assert str(obj) == expected_str


@pytest.mark.parametrize('pkg_dict, expected_exception', [
    ({}, ValueError),
    ({'package': 'x'}, ValueError),
    ({'architecture': 'x'}, ValueError),
    ({'version': 'x'}, ValueError),
    ({'package': None, 'architecture': None, 'version': None}, ValueError),
    ({'package': None, 'architecture': 'x',  'version': 'x'}, ValueError),
    ({'package': 'x',  'architecture': None, 'version': 'x'}, ValueError),
    ({'package': 'x',  'architecture': 'x',  'version': None}, ValueError),
])
def test_object_error(pkg_dict, expected_exception):
    """Checks that BinaryPackage objects can be instantiated."""
    with pytest.raises(expected_exception):
        obj = BinaryPackage(pkg_dict)
        assert obj


@pytest.mark.parametrize('pkg_dict, expected_binary_package_name', [
    ({'package': 'a',  'architecture': 'b',  'version': '1'}, 'a'),
])
def test_name(pkg_dict, expected_binary_package_name):
    """Checks the package name is obtained from pkg_dict."""
    binary_package = BinaryPackage(pkg_dict)

    assert binary_package.name == expected_binary_package_name


@pytest.mark.parametrize('pkg_info, expected_deps', [
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'depends': 'a'},
     {'a': ''}),
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'depends': 'a, b, c'},
     {'a': '', 'b': '', 'c': ''}),
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'depends': 'a | b | c'},
     {('a', 'b', 'c'): {'a': '', 'b': '', 'c': ''}}),
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'depends': 'a (= 1) | b (> 2) | c (<= 3)'},
     {('a', 'b', 'c'): {'a': '(= 1)', 'b': '(> 2)', 'c': '(<= 3)'}}),
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'depends': 'a (>= 1.0)'},
     {'a': '(>= 1.0)'}),
    ({'package': 'zlib1g-dev',  'architecture': 'amd64',  'version': '1:1.2.13.dfsg-1ubuntu4',
      'depends': 'zlib1g (= 1:1.2.13.dfsg-1ubuntu4), libc6-dev | libc-dev'},
     {('libc6-dev', 'libc-dev'): {'libc-dev': '', 'libc6-dev': ''},
      'zlib1g': '(= 1:1.2.13.dfsg-1ubuntu4)'}),
])
def test_installation_dependencies(pkg_info, expected_deps):
    """Checks that BinaryPackage objects parse their Dependencies field."""
    binary_package = BinaryPackage(pkg_info)

    assert binary_package.installation_dependencies == expected_deps


@pytest.mark.parametrize('pkg_info, expected_recommends', [
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'recommends': 'a'},
     {'a': ''}),
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'recommends': 'a'},
     {'a': ''}),
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'recommends': 'a, b, c'},
     {'a': '', 'b': '', 'c': ''}),
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'recommends': 'a | b | c'},
     {('a', 'b', 'c'): {'a': '', 'b': '', 'c': ''}}),
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'recommends': 'a (= 1) | b (> 2) | c (<= 3)'},
     {('a', 'b', 'c'): {'a': '(= 1)', 'b': '(> 2)', 'c': '(<= 3)'}}),
    ({'package': 'x',  'architecture': 'x',  'version': 'x', 'recommends': 'a (>= 1.0)'},
     {'a': '(>= 1.0)'}),
    ({'package': 'zlib1g-dev',  'architecture': 'amd64',  'version': '1:1.2.13.dfsg-1ubuntu4',
      'recommends': 'zlib1g (= 1:1.2.13.dfsg-1ubuntu4), libc6-dev | libc-dev'},
     {('libc6-dev', 'libc-dev'): {'libc-dev': '', 'libc6-dev': ''},
      'zlib1g': '(= 1:1.2.13.dfsg-1ubuntu4)'}),
])
def test_recommended_packages(pkg_info, expected_recommends):
    """Checks that BinaryPackage objects parse their Recommends field."""
    binary_package = BinaryPackage(pkg_info)

    assert binary_package.recommended_packages == expected_recommends



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2019-2020 Bryce Harrington <bryce@canonical.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# pylint: disable=W0212

"""Tests the Launchpad class as an interface to the Launchpad service API.

Tests for our launchpadlib wrapper & helper methods.
"""

import sys
import os.path
from unittest.mock import Mock, patch

import logging
import pytest
from launchpadlib.launchpad import Launchpad

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.lp import Lp

APPNAME = 'test-lp'


@pytest.fixture(name='fakelp')
def fixture_fakelp():
    """Connect to Launchpad."""
    mock_launchpad = Mock(spec=Launchpad)

    return Lp(
        application_name=APPNAME,
        service=mock_launchpad)


def test_new_connection():
    """Verifies the Lp object will auto-login properly."""
    mock_launchpad = Mock(spec=Launchpad)

    lp = Lp(
        application_name=APPNAME,
        service=mock_launchpad)

    # Cause the lp._instance() internal property to be triggered
    logging.debug(lp.me)
    logging.debug(lp.people)

    # Verify the triggering resulted in a login attempt
    mock_launchpad.login_with.assert_called_once_with(
        application_name=APPNAME,
        service_root='production',
        allow_access_levels=['WRITE_PRIVATE'],
        version='devel',
    )


@patch("ppa.lp.Credentials")
def test_new_connection_creds(credentials_mock):
    """Verifies the Lp object will login if credentials are provided."""
    mock_launchpad = Mock(spec=Launchpad)

    lp = Lp(
        application_name=APPNAME,
        service=mock_launchpad,
        credentials='...')

    # Cause the lp._instance() internal property to be triggered
    logging.debug(lp.me)
    logging.debug(lp.people)

    # Verify the triggering did not result in a login attempt
    mock_launchpad.login_with.assert_not_called()
    # Rather that the login was done via envvar
    # creating the credentials
    credentials_mock.from_string.assert_called_once()
    # and the service was created with the given credentials
    mock_launchpad.assert_called_once_with(
        credentials_mock.from_string.return_value,
        None,
        None,
        service_root="production",
        version="devel"
    )


def test_api_root(fakelp):
    """Ensures we can get LP's API root."""
    fakelp.load(Lp.API_ROOT_URL + 'hi')
    fakelp._instance.load.assert_called_once_with(
        'https://api.launchpad.net/devel/hi')


def test_ubuntu(fakelp):
    """Checks the .ubuntu property."""
    fakelp._instance.distributions = {'ubuntu': 'UBUNTU'}
    assert fakelp.ubuntu == 'UBUNTU'


def test_ubuntu_active_series(fakelp):
    """Checks the active series names for Ubuntu."""
    mock_hirsute = Mock(active=False)
    mock_hirsute.name = 'hirsute'

    mock_jammy = Mock(active=True)
    mock_jammy.name = 'jammy'

    mock_ubuntu = Mock(series=[mock_hirsute, mock_jammy])
    fakelp._instance.distributions = {'ubuntu': mock_ubuntu}
    assert fakelp.ubuntu_active_series()[0] == mock_jammy
    assert fakelp.ubuntu_active_series()[0].name == 'jammy'


def test_ubuntu_devel_series(fakelp):
    """Checks the devel series name for Ubuntu."""
    mock_mantic = Mock(active=False)
    mock_mantic.name = 'mantic'

    mock_jammy = Mock(active=True)
    mock_jammy.name = 'jammy'

    mock_ubuntu = Mock(series=[mock_mantic, mock_jammy])
    mock_ubuntu.getDevelopmentSeries.return_value = [mock_mantic]
    fakelp._instance.distributions = {'ubuntu': mock_ubuntu}
    assert fakelp.ubuntu_devel_series() == [mock_mantic]
    assert fakelp.ubuntu_devel_series()[0].name == 'mantic'


def test_ubuntu_stable_series(fakelp):
    """Checks the stable series name for Ubuntu."""
    mock_hirsute = Mock(active=False)
    mock_hirsute.name = 'hirsute'

    mock_jammy = Mock(status='Supported')
    mock_jammy.name = 'jammy'

    mock_kinetic = Mock(status='Obsolete')
    mock_kinetic.name = 'kinetic'

    mock_lunar = Mock(status='Current Stable Release')
    mock_lunar.name = 'lunar'

    mock_mantic = Mock(status='Active Development')
    mock_mantic.name = 'mantic'

    mock_mantic = Mock(status='Future')
    mock_mantic.name = 'nseries'

    mock_ubuntu = Mock(series=[mock_hirsute, mock_jammy, mock_kinetic, mock_lunar, mock_mantic])
    fakelp._instance.distributions = {'ubuntu': mock_ubuntu}
    assert fakelp.ubuntu_stable_series() == [mock_lunar]


def test_debian(fakelp):
    """Checks the .debian property."""
    fakelp._instance.distributions = {'debian': 'DEBIAN'}
    assert fakelp.debian == 'DEBIAN'


def test_debian_active_series(fakelp):
    """Checks the active series names for Debian."""
    mock_woody = Mock(active=False)
    mock_woody.name = 'woody'

    mock_buster = Mock(active=True)
    mock_buster.name = 'buster'

    mock_sid = Mock(active=False)
    mock_sid.name = 'experimental'

    mock_debian = Mock(series=[mock_woody, mock_buster, mock_sid])
    fakelp._instance.distributions = {'debian': mock_debian}
    assert fakelp.debian_active_series()[0].name == 'buster'


def test_debian_experimental_series(fakelp):
    """Checks the experimental series name for Debian."""
    mock_woody = Mock(active=False)
    mock_woody.name = 'woody'

    mock_buster = Mock(active=True)
    mock_buster.name = 'buster'

    mock_sid = Mock(active=False)
    mock_sid.name = 'experimental'

    mock_debian = Mock(series=[mock_woody, mock_buster, mock_sid])
    fakelp._instance.distributions = {'debian': mock_debian}
    assert fakelp.debian_experimental_series() == mock_sid


def test_get_teams(fakelp):
    """Checks that a person's teams can be retrieved."""
    mock_me = Mock(memberships_details=[
        Mock(name='t1', self_link="https://launchpad.net/~t1"),
        Mock(name='t2', self_link="https://launchpad.net/~t2/+blah"),
        Mock(name='t3', self_link="https://launchpad.net/~t3/+blah"),
    ])
    fakelp._instance.people = {'me': mock_me}
    assert fakelp.get_teams('me') == ['t1', 't2', 't3']



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2022-2024 Authors
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

"""A directive to run a DEP8 test against a source package."""

from functools import lru_cache
from urllib.parse import urlencode
from typing import List

from .constants import URL_AUTOPKGTEST
from .text import ansi_hyperlink
from .source_publication import SourcePublication


class Trigger:
    """A publication and architecture to use when running an autopkgtest.

    A trigger indicates a source package whose autopkgtest(s) to invoke
    for a given hardware architecture, after installing a particular
    version of that package, and possibly other source packages, from a
    given series.

    A Job can have multiple Triggers, each against a different source
    package and/or architectures, but all such Triggers must be against
    the same series as the Job itself.
    """
    def __init__(self, source_publication, architecture, test_package=None):
        """Initialize a new Trigger for a given publication and architecture.

        :param SourcePublication source_publication: The source package
            publication object, which includes its name, version,
            series, and archive.
        :param str architecture: The hardware architecture to run on.
        :param str test_package: The package to run autopkgtests from.
        """
        self.source_publication = source_publication
        self.architecture = architecture
        if test_package:
            self.test_package = test_package
        else:
            self.test_package = source_publication.package

    def __repr__(self) -> str:
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (f'{self.__class__.__name__}('
                f'source_publication={self.source_publication!r}, '
                f'architecture={self.architecture!r}, '
                f'test_package={self.test_package!r})')

    def __str__(self) -> str:
        """Return a human-readable summary of the object.

        :rtype: str
        :returns: Printable summary of the object.
        """
        return f"{self.test_package}: {self.source_publication} [{self.architecture}]"

    @lru_cache
    def to_dict(self) -> dict:
        """Return a basic dict structure of the Trigger's data."""
        return {
            'source_publication': self.source_publication.to_dict(),
            'architecture': self.architecture,
            'test_package': self.test_package
        }

    @property
    @lru_cache
    def action_url(self) -> str:
        """Renders the trigger as a URL to start running the test.

        :rtype: str
        :returns: Trigger action URL
        """
        params = [
            ("release", self.source_publication.series),
            ("package", self.test_package),
            ("arch", self.architecture),
        ]

        # Trigger for the source package itself
        params.append(
            (
                "trigger",
                f"{self.source_publication.package}/{self.source_publication.version}")
            )

        # TODO: Additional triggers...

        # The PPA, if one is defined for this trigger
        if self.source_publication.archive:
            params.append(("ppa", self.source_publication.archive))

        return f"{URL_AUTOPKGTEST}/request.cgi?" + urlencode(params)


# TODO: Use SourcePub?
def get_triggers(package, version, ppa, series, architectures,
                 sources=None) -> List[Trigger]:
    """Returns Triggers for the given criteria.

    :param str package: The source package name.
    :param str version: The version of the source package to install.
    :param Ppa ppa: Archive containing the package to run tests against.
    :param str series: The distro release series codename.
    :param list[str] architectures: The architectures to provide triggers for.
    :param list[str] sources: (Unimplemented)
    :rtype: list[Trigger]
    :returns: List of triggers, if any, or an empty list on error.
    """
    source_pub = SourcePublication(package, version, series, archive=ppa)
    return [
        Trigger(source_pub, arch, sources)
        for arch
        in architectures
    ]


# TODO: Use SourcePub?
def show_triggers(triggers,
                  show_trigger_urls=False,
                  show_trigger_names=False):
    """Prints a list of triggers.

    :param list[Trigger] triggers:  The triggers to be displayed.
    :param str status: Result of the triggered test run.
    :param bool show_trigger_urls: If true, print out the trigger URLs
        as text; otherwise triggers will be printed as a hyperlink named
        'package/version'.  This is necessary for terminals lacking ANSI
        hyperlink support, for example.
    :param bool show_trigger_names: If true, includes display of the
        package names for triggers.  This may be useful if printing
        complex triggers or triggers for multiple different packages.
    """
    if show_trigger_urls:
        for trigger in triggers:
            title = ''
            arch = trigger.architecture
            if show_trigger_names:
                title = trigger.test_package
            print(f"    + {title}@{arch}: {trigger.action_url} ♻️ ")
        for trigger in triggers:
            title = ''
            arch = trigger.architecture
            if show_trigger_names:
                title = trigger.test_package
            url = trigger.action_url + "&all-proposed=1"
            print(f"    + {title}@{arch}: {url} 💍")

    else:
        for trigger in triggers:
            title = ''
            if show_trigger_names:
                title = trigger.test_package
            arch = trigger.architecture
            pad = ' ' * (1 + abs(len('ppc64el') - len(arch)))

            basic_trig = ansi_hyperlink(
                trigger.action_url, f"Trigger basic {title}@{arch}♻️ "
            )
            all_proposed_trig = ansi_hyperlink(
                trigger.action_url + "&all-proposed=1",
                f"Trigger all-proposed {title}@{arch}💍"
            )
            print("    + " + pad.join([basic_trig, all_proposed_trig]))


if __name__ == "__main__":
    import json

    print('##############################')
    print('## Trigger class smoke test ##')
    print('##############################')
    print()

    print("Basic trigger")
    print("-------------")
    pub = SourcePublication('my-package', '1.2.3', 'kinetic')
    t = Trigger(pub, 'amd64')
    print(pub)
    print(t)
    print(t.source_publication.autopkgtest_history_url)
    print(t.action_url)
    print()

    print("Object Dump")
    print("-----------")
    t = Trigger(pub, 'i386', 'ppa:me/myppa')
    print(json.dumps(t.to_dict(), indent=4))
    print()

    print("* PPA trigger:")
    pub = SourcePublication('my-ppa-package', '3.2.1', 'kinetic', 'my-ppa')
    t = Trigger(pub, 'amd64')
    print(t)
    print(t.source_publication.autopkgtest_history_url)
    print(t.action_url)
    print()



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2019 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests the PpaGroup class as a way to manage a Team or Person's PPAs."""

import os
import sys

import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.ppa_group import PpaGroup, PpaAlreadyExists
from tests.helpers import LpServiceMock


def test_object():
    """Checks that PpaGroup objects can be instantiated."""
    ppa_group = PpaGroup(service=LpServiceMock(), name='test-ppa')
    assert ppa_group

    with pytest.raises(ValueError):
        ppa_group = PpaGroup(service=LpServiceMock(), name=None)
    with pytest.raises(ValueError):
        ppa_group = PpaGroup(service=None, name='test-ppa')


def test_create_ppa():
    """Checks that PpaGroups can create PPAs."""
    name = 'test_ppa'
    ppa_group = PpaGroup(service=LpServiceMock(), name='me')
    ppa = ppa_group.create(name)
    assert ppa is not None
    assert name in ppa.address
    assert type(ppa.description) is str


def test_create_existing_ppa():
    """Check exception creating an already created PPA."""
    name = 'test_ppa'
    ppa_group = PpaGroup(service=LpServiceMock(), name='me')
    ppa_group.create(name)
    with pytest.raises(PpaAlreadyExists):
        ppa_group.create(name)


def test_create_with_description():
    """Check setting a description for a PPA."""
    ppa_group = PpaGroup(service=LpServiceMock(), name='me')
    description = 'PPA Test Description'
    ppa = ppa_group.create('test_ppa_with_description', description)
    assert ppa is not None
    assert ppa.description == description


def test_create_with_owner():
    """Check creating a PPA for a particular owner."""
    lp = LpServiceMock()
    lp.launchpad.add_person('test_owner_name')
    ppa_group = PpaGroup(service=lp, name='test_owner_name')
    ppa = ppa_group.create('ppa_test_name')
    assert ppa is not None
    assert ppa.address == 'ppa:test_owner_name/ppa_test_name'


def test_create_private():
    """Check creating a private PPA."""
    lp = LpServiceMock()
    ppa_group = PpaGroup(service=lp, name='me')
    ppa = ppa_group.create('private_ppa', private=True)
    assert ppa is not None
    assert ppa.address == 'ppa:me/private_ppa'
    assert ppa.is_private is True


def test_list_ppas():
    """Check listing the PPAs for a PPA group."""
    test_ppa_list = ['a', 'b', 'c', 'd']
    ppa_group = PpaGroup(service=LpServiceMock(), name='me')

    # Add several ppas
    for ppa in test_ppa_list:
        ppa_group.create(ppa)

    ppas = [ppa.name for ppa in list(ppa_group.ppas)]
    assert test_ppa_list == ppas



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2022 Authors
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

"""The completed data from an autopkgtest run."""

import re
import urllib.request
from functools import lru_cache
from typing import Dict, List, Iterator
import gzip
import time

from .source_publication import SourcePublication
from .subtest import Subtest
from .trigger import Trigger
from .text import ansi_hyperlink


class Result:
    """The completed data from an autopkgtest run Job.

    This object provides access to the test run's settings and results.
    """
    VALUES = {
        'PASS': "✅",
        'FAIL': "❌",
        'BAD': "⛔"
    }

    def __init__(self, url, time, series, arch, source):
        """Initialize a new Result object.

        :param str url: HTTP path to the test log for this result.
        :param str time: The execution time of the test run.
        :param str series: The distro release series codename.
        :param str arch: The architecture for the result.
        :param str source:
        """
        self.url = url
        self.time = time
        self.series = series
        self.arch = arch
        self.source = source
        self.error_message = None
        self._log = None

    def __repr__(self) -> str:
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (f'{self.__class__.__name__}('
                f'url={self.url!r}, '
                f'time={self.time!r}, '
                f'series={self.series!r}, '
                f'arch={self.arch!r}, '
                f'source={self.source!r})')

    def __str__(self) -> str:
        """Return a human-readable summary of the object.

        :rtype: str
        :returns: Printable summary of the object.
        """
        pad = ' ' * (1 + abs(len('ppc64el') - len(self.arch)))
        return f"{self.source} on {self.series} for {self.arch}{pad}@ {self.timestamp}"

    @lru_cache
    def to_bullet_tree(self, show_urls=False):
        """Returns results in a bulletted list.

        :param bool show_urls: If true, displays URLs as plain text
            strings; if False, it generates a clickable ANSI hyperlink for
            the log.
        :returns: Human-readable results with logs in bulletted list.
        :rtype: str
        """
        output = ''
        if show_urls:
            log_str = f"      • Log: {self.url}"
            output = f"    + {self.status_icon} {self}\n{log_str}"
        else:
            log_link = ansi_hyperlink(self.url, "Log️ 🗒️ ")
            output = f"    + {self.status_icon} {self}  {log_link}"
        output += "\n"
        if self.status != 'PASS':
            if show_urls:
                output += f"      • Status: {self.status}\n"
            for subtest in self.get_subtests():
                output += f"      • {subtest}\n"
        return output

    @lru_cache
    def to_dict(self) -> Dict:
        """Return a basic dict structure of the Result's data."""
        return {
            'url': self.url,
            'timestamp': self.timestamp,
            'series': self.series,
            'arch': self.arch,
            'source': self.source,
            'error_message': self.error_message,
            'log': self.log,
            'triggers': [tr.to_dict() for tr in self.get_triggers()],
            'subtests': [st.to_dict() for st in self.get_subtests()],
            'status': self.status,
            'status_icon': self.status_icon
        }

    @property
    def timestamp(self) -> str:
        """The result's completion time formatted as a string."""
        return time.strftime("%d.%m.%y %H:%M:%S", self.time)

    @property
    @lru_cache
    def log(self) -> str:
        """The log contents for results, downloading if necessary.

        Retrieves the log via the result url, handles decompression, and
        caches the results internally, so that subsequent calls don't
        re-download the data.

        On error, returns None and stores the error message in
        the Result.error_message property.

        :rtype: str
        :returns: Full text of the log file, or None on error.
        """
        request = urllib.request.Request(self.url)
        request.add_header('Cache-Control', 'max-age=0')
        try:
            response = urllib.request.urlopen(request)
        except urllib.error.HTTPError as e:
            self.error_message = f"Failed to Download Test Log ⚪: {e}"
            return None

        result_gzip = response.read()
        try:
            return gzip.decompress(result_gzip).decode("utf-8",
                                                       errors="replace")
        except UnicodeDecodeError:
            self.error_message = "Broken Test Log ⚪"
            return None

    # TODO: Merge triggers and get_triggers()
    @property
    @lru_cache
    def triggers(self) -> List[str]:
        """The package/version parameters used to generate this Result.

        This returns the set of triggers used to create the Result, as
        recorded in the test log.  Each trigger is a package/version
        pair corresponding to source packages to use from the proposed
        archive (instead of from the release archive).

        :rtype: List[str]
        :returns: List of package/version triggers.
        """
        regex_triggers = re.compile(r'--env=ADT_TEST_TRIGGERS=(.*?) -- ')
        header_split = self.log.split(": @@@@@@@@@@@@@@@@@@@@", 1)
        m = re.search(regex_triggers, header_split[0])
        if not m:
            return []
        return m.group(1).strip("'").split()

    @lru_cache
    def get_triggers(self, name=None) -> Iterator[Trigger]:
        """Return list of Triggers that were used to create this Result.

        This returns the set of Triggers used to create the Result, as
        recorded in the test log.  Each trigger identifies a
        package/version pair corresponding to source packages to use
        from the proposed archive (instead of from the release archive).

        :param str name: If defined, only return triggers starting with this name.
        :rtype: Iterator[Trigger]
        :returns: Triggers used to generate this Result, if any, or an empty list
        """
        if not self.triggers:
            return []

        for trigger in self.triggers:
            pkg, ver = trigger.split('/', 1)
            yield Trigger(
                SourcePublication(
                    package=pkg,
                    version=ver,
                    series=self.series
                ),
                architecture=self.arch,
            )

    @lru_cache
    def get_subtests(self, name=None) -> List[Subtest]:
        """Return list of Subtests that were run for this Result.

        :param str name: Only display subtests starting with this.
        :rtype: List[Subtest]
        :returns: Subtests completed for this Result, or empty list.
        """
        result_split = self.log.split("@@@@@@@@@@@@@@@@@@@@ summary", 1)
        if len(result_split) < 2:
            self.error_message = "Failure setting up testbed ⚪"
            return [Subtest("testbed setup failure  BAD")]

        subtests = []
        result_sum = result_split[1]
        for line in re.findall("(.*PASS|.*SKIP|.*FAIL|.*FLAKY|.*BAD)", result_sum):
            if re.match(r"^[0-9]+s\s", line):
                # Newer autopkgtest logs are prefixed with the timestamp.  Ignore it.
                line = line.split(' ', 1)[1]
            if name and not line.startswith(name):
                continue
            subtests.append(Subtest(line))
        return subtests

    @property
    @lru_cache
    def status(self) -> str:
        """The overall status of all subtests.

        If the triggered run completed successfully, then the status will
        be either FAIL if any of the subtests failed, or PASS otherwise.

        If the run did not complete successfully, then a 'BAD' status
        will be returned, and the reason can be examined via the
        Result.error_message property.

        :rtype: str
        :returns: 'PASS', 'FAIL', or 'BAD'
        """
        if self.error_message:
            return 'BAD'

        for subtest in self.get_subtests():
            if subtest.status == 'FAIL':
                return 'FAIL'
            elif subtest.status == 'BAD':
                return 'BAD'
        return 'PASS'

    @property
    @lru_cache
    def status_icon(self) -> str:
        """A unicode symbol corresponding to test's overall status.

        :rtype: str
        :returns: Unicode symbol
        """
        return Result.VALUES[self.status]


def get_results(response, base_url, arches=None, sources=None) -> Iterator[Result]:
    """Return iterator of Results from the base URL for given criteria.

    Retrieves the autopkgtest results limited to the given architectures
    and source packages.  If unspecified, returns all results.

    :param str base_url: URL for the autopkgtest results.
    :param List[str] arches: Architectures to include in results.
    :param List[str] sources: Source packages to include in results.
    :rtype: Iterator[Result]
    :returns: Iterable results, if any, or an empty list on error
    """
    if response is None:
        return []
    for line in response.read().split(b'\n'):
        if line == b'' or not line.endswith(b"log.gz"):
            continue
        result = line.decode("utf-8")
        series, arch, _, source, timestamp = result.split('/')[:5]
        if (arches and (arch not in arches)):
            continue
        if (sources and (source not in sources)):
            continue
        if not base_url.endswith('/') and not result.startswith('/'):
            base_url += '/'
        yield Result(
            url=base_url + result,
            time=time.strptime(timestamp[:-7], "%Y%m%d_%H%M%S"),
            series=series,
            arch=arch,
            source=source)


def show_results(results: List[tuple], show_urls: bool = False):
    """Prints the results of completed tests.

    :param List[tuple] results: The test results to be displayed.
    :param bool show_urls: If true, print textual http links instead of
        clickable ANSI hyperlinks.
    """
    n = 0
    for n, trigger_sets in enumerate(results, start=1):
        if n == 1:
            print("* Results:")
        for trigger, trigger_results in trigger_sets.items():
            trigger_result_str = ''
            for trigger_result in trigger_results:
                trigger_result_str += trigger_result.to_bullet_tree(show_urls)
            print(f"  - {trigger}\n{trigger_result_str.rstrip()}")
    if n == 0:
        print("* Results: (none)")


if __name__ == "__main__":
    import os
    import json

    from ppa.io import open_url
    from ppa.constants import ARCHES_AUTOPKGTEST, URL_AUTOPKGTEST

    print('#############################')
    print('## Result class smoke test ##')
    print('#############################')
    print()

    print("Basic result")
    print("------------")
    timestamp = time.strptime('20030201_040506', "%Y%m%d_%H%M%S")
    result_1 = Result('url-here', timestamp, 'kinetic', 'amd64', 'my-package')
    print("* Result object:")
    print(repr(result_1))
    print(result_1)
    print()

    data_dir = os.path.realpath(os.path.join(os.path.dirname(__file__), "../tests/data"))
    url = f"file://{data_dir}/results-six-s390x.log.gz"
    result_2 = Result(url, timestamp, 'kinetic', 'amd64', 'my-package')
    print("* Log Head:")
    print("\n".join(result_2.log.splitlines()[0:4]))
    print()

    # TODO: Implement something that dumps the passing tests for given package from -proposed
    # TODO: Filter to items with only Pass, Not a regression, or No test results

    print("Object Dump")
    print("-----------")
    result_3 = Result(url, timestamp, 'kinetic', 'amd64', 'my-package')
    print(json.dumps(result_3.to_dict(), indent=4))
    print()

    print("Loading live excuses data")
    print("-------------------------")
    base_results_fmt = f"{URL_AUTOPKGTEST}/results/autopkgtest-%s-%s-%s/"
    base_results_url = base_results_fmt % ('mantic', 'bryce', 'apache2-merge-v2.4.54-3')
    url = f"{base_results_url}?format=plain"
    response = open_url(url)

    for result in get_results(response, base_results_url, arches=ARCHES_AUTOPKGTEST):
        print(f"* {result}")
        print("  - Triggers: " + ', '.join([str(r) for r in result.get_triggers()]))

        for subtest in result.get_subtests():
            print(f"  - {subtest}")

        print()



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2012 Bryce Harrington <bryce@bryceharrington.org>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Routines to encode or convert to and from text.'''

import sys
from decimal import Decimal
from functools import lru_cache


# Constants for ANSI codes
OSC8 = "\u001b]8"
ST = "\u001b\\"


@lru_cache
def quote(msg):
    """Similar to urllib.quote but for glibs' GMarkup.

    @param msg: string to quote
    @returns: quoted string
    """
    msg = msg.replace('&', '&amp;')
    msg = msg.replace('<', '&lt;')
    msg = msg.replace('>', '&gt;')
    return msg


@lru_cache
def o2str(obj):
    """Return string representation of various types of objects.

    Convert a unicode, decimal.Decimal, datetime object, etc. to a str.
    Converts lists and tuples of objects into lists of strings.
    """
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, bytes):
        return obj.decode("utf-8")
    elif isinstance(obj, Decimal):
        return str(obj)
    elif isinstance(obj, list) or isinstance(obj, tuple):
        new_list = []
        for item in obj:
            new_list.append(o2str(item))
        return new_list
    elif str(type(obj)) == "<type 'datetime.datetime'>":
        return obj.ctime()
    else:
        # print str(type(obj))
        return obj


@lru_cache
def to_bool(value):
    """Convert 'something' to boolean. Raises exception for invalid formats.

    Possible True  values: 1, True, '1', 'TRue', 'yes', 'y', 't'
    Possible False values: 0, False, None, [], {}, '', '0', 'faLse', 'no', 'n', 'f', 0.0
    """
    if isinstance(value, str):
        if value.lower() in ("yes", "y", "true", "t", "1"):
            return True
        if value.lower() in ("no", "n", "false", "f", "0", "none", ""):
            return False
        raise Exception('Invalid value for boolean conversion: ' + value)
    return bool(value)


@lru_cache
def o2float(value):
    '''Convert strings like 42%, 123M, 1.2B into floating point numbers.

    Returned values are in millions, so '1.2B' returns 1200
    '''
    if value is None:
        return 0.0
    elif type(value) is float:
        return value
    elif type(value) is int:
        return float(value)
    elif value == '--':
        return 0.0

    value = value.replace(',', '')
    last = value[len(value) - 1]
    if last == 'M':
        return float(value[:-1])
    elif last == 'B':
        return float(value[:-1]) * 1000
    elif last == '%':
        return float(value[:-1]) / 100.0
    elif last == ')' and value[0] == '(':
        return -1 * o2float(value[1:-1])

    try:
        return float(value)
    except ValueError:
        sys.stderr.write(f"ofloat: Could not convert '{value}' to float\n")
        raise


def ansi_hyperlink(url, text):
    """Format text into a hyperlink using ANSI escape codes.

    :ref: https://iterm2.com/documentation-escape-codes.html
    :ref: https://gist.github.com/egmontkob/eb114294efbcd5adb1944c9f3cb5feda

    :param str url: The linked action to load in a web browser.
    :param str text: The visible text to show.
    :rtype: str
    :returns: ANSI escape code sequence to display the hyperlink.
    """
    return f"{OSC8};;{url}{ST}{text}{OSC8};;{ST}"


if __name__ == "__main__":
    print('#####################')
    print('## Text smoke test ##')
    print('#####################')
    print()

    print(ansi_hyperlink("https://launchpad.net/ppa-dev-tools", "ppa-dev-tools"))



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2022 Authors
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

"""Utilities for reading input and writing output to external locations."""

import sys
import urllib.request


def open_url(url, desc="data"):
    """Open a remote URL for reading.

    :rtype: urllib.request.Request
    :returns: A request object for the stream to read from, or None on error.
    """
    request = urllib.request.Request(url)
    request.add_header('Cache-Control', 'max-age=0')
    try:
        return urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        if e.code == 401:
            # 401 means access denied.  Launchpad sometimes returns this
            # when something is not published yet.
            return None
        elif e.code == 404:
            # 404 means not found; prefer not to emit error messages
            return None
        else:
            sys.stderr.write(f"Error: Could not retrieve {desc} from {url}: {e}\n")
            return None



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2022 Authors
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

"""An individual DEP8 test run."""

import json
from functools import lru_cache


class Subtest:
    """An autopkgtest sub-component test run.

    A triggered autopkgtest can invoke multiple DEP8 tests, such as running
    checks on dependencies, the software's testsuite, and integration tests.
    Each of these is considered a "Subtest"
    """
    VALUES = {
        'PASS': "🟩",
        'SKIP': "🟧",
        'FAIL': "🟥",
        'FLAKY': "🟫",
        'BAD': "⛔",
        'UNKNOWN': "⚪"
    }

    def __init__(self, line):
        """Initialize a new Subtext object.

        :param str line: The subtest result summary from a test log.
        """
        if not line:
            raise ValueError("undefined line.")

        self._line = line

    def __repr__(self) -> str:
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (f'{self.__class__.__name__}('
                f'line={self._line!r})')

    def __str__(self) -> str:
        """Return a human-readable summary of the object.

        :rtype: str
        :returns: Printable summary of the object.
        """
        return f"{self.desc:25} {self.status:6} {self.status_icon}"

    @lru_cache
    def to_dict(self) -> dict:
        """Return a basic dict structure of the Subtest's data."""
        return {
            'line': self._line,
            'desc': self.desc,
            'status': self.status,
            'status_icon': self.status_icon
        }

    @property
    @lru_cache
    def desc(self) -> str:
        """The descriptive text for the given subtest.

        :rtype: str
        :returns: Descriptive text.
        """
        return next(iter(self._line.split()), '')

    @property
    @lru_cache
    def status(self) -> str:
        """The success or failure of the given subtest.

        :rtype: str
        :returns: Status term in capitalized letters (PASS, FAIL, etc.)
        """
        for k in Subtest.VALUES:
            if f" {k}" in self._line:
                return k
        return 'UNKNOWN'

    @property
    @lru_cache
    def status_icon(self) -> str:
        """A unicode symbol corresponding to subtest's status.

        :rtype: str
        :returns: Single unicode character matching the status.
        """
        return Subtest.VALUES[self.status]


if __name__ == "__main__":
    print('##############################')
    print('## Subtest class smoke test ##')
    print('##############################')
    print()

    print("Valid cases")
    print("-----------")
    print(Subtest('subtest-a   UNKNOWN'))
    print(Subtest('subtest-b   PASS'))
    print(Subtest('subtest-c   FAIL'))
    print(Subtest('subtest-d   FAIL non-zero exit status 123'))
    print(Subtest('subtest-e   SKIP'))
    print(Subtest('subtest-f   BAD'))
    print(Subtest('librust-clang-sys-dev:clang_10_0 FAIL non-zero exit status 101'))
    print(Subtest('librust-clang-sys-dev:static FLAKY non-zero exit status 101'))
    print()

    print("Invalid cases")
    print("-------------")
    print(Subtest('invalid subtest: invalid'))
    print(Subtest('bAd subtest: bAd'))
    print()

    print("Object Dump")
    print("-----------")
    s = Subtest('librust-clang-sys-dev:clang_10_0 FAIL non-zero exit status 101')
    print(json.dumps(s.to_dict(), indent=4))
    print()
